"""
Client classes for the GA4GH reference implementation.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import requests
import posixpath
import logging

import ga4gh.protocol as protocol
import ga4gh.pb as pb
import ga4gh.exceptions as exceptions


class AbstractClient(object):
    """
    The abstract superclass of GA4GH Client objects.
    """

    def __init__(self, log_level=0):
        self._page_size = None
        self._log_level = log_level
        self._protocol_bytes_received = 0
        logging.basicConfig()
        self._logger = logging.getLogger(__name__)
        self._logger.setLevel(log_level)

    def _deserialize_response(
            self, json_response_string, protocol_response_class):
        self._protocol_bytes_received += len(json_response_string)
        self._logger.debug("response:{}".format(json_response_string))
        if not json_response_string:
            raise exceptions.EmptyResponseException()
        return protocol.fromJson(json_response_string, protocol_response_class)

    def _run_search_page_request(
            self, protocol_request, object_name, protocol_response_class):
        """
        Runs a complete transaction with the server to obtain a single
        page of search results.
        """
        raise NotImplemented()

    def _run_search_request(
            self, protocol_request, object_name, protocol_response_class):
        """
        Runs the specified request at the specified object_name and
        instantiates an object of the specified class. We yield each object in
        listAttr.  If pages of results are present, repeat this process
        until the pageToken is null.
        """
        not_done = True
        while not_done:
            response_object = self._run_search_page_request(
                protocol_request, object_name, protocol_response_class)
            value_list = getattr(
                response_object,
                protocol.getValueListName(protocol_response_class))
            for extract in value_list:
                yield extract
            not_done = bool(response_object.next_page_token)
            protocol_request.page_token = response_object.next_page_token

    def _run_list_reference_bases_page_request(self, protocol_request):
        """
        Runs a complete transaction with the server to get a single
        page of results for the specified ListReferenceBasesRequest.
        """
        raise NotImplemented()

    def list_reference_bases(self, id_, start=0, end=None):
        """
        Returns an iterator over the bases from the server in the form
        of consecutive strings. This command does not conform to the
        patterns of the other search and get requests, and is implemented
        differently.
        """
        request = protocol.ListReferenceBasesRequest()
        request.start = pb.int(start)
        request.end = pb.int(end)
        request.reference_id = id_
        not_done = True
        # TODO We should probably use a StringIO here to make string buffering
        # a bit more efficient.
        bases_list = []
        while not_done:
            response = self._run_list_reference_bases_page_request(request)
            bases_list.append(response.sequence)
            not_done = bool(response.next_page_token)
            request.page_token = response.next_page_token
        return "".join(bases_list)

    def _run_get_request(self, object_name, protocol_response_class, id_):
        """
        Requests an object from the server and returns the object of
        type protocol_response_class that has id id_.
        Used for requests where a single object is the expected response.
        """
        raise NotImplemented()

    def get_bio_sample(self, bio_sample_id):
        """
        Perform a get request for the given BioSample.

        :param str bio_sample_id: The ID of the BioSample
        :return: The requested BioSample.
        :rtype: :class:`ga4gh.protocol.BioSample`
        """
        return self._run_get_request(
            "biosamples", protocol.BioSample, bio_sample_id)

    def get_individual(self, individual_id):
        """
        Perform a get request for the given Individual.

        :param str individual_id: The ID of the Individual
        :return: The requested Individual.
        :rtype: :class:`ga4gh.protocol.Individual`
        """
        return self._run_get_request(
            "individuals", protocol.Individual, individual_id)

    def get_page_size(self):
        """
        Returns the suggested maximum size of pages of results returned by
        the server.
        """
        return self._page_size

    def set_page_size(self, page_size):
        """
        Sets the requested maximum size of pages of results returned by the
        server to the specified value.
        """
        self._page_size = page_size

    def get_protocol_bytes_received(self):
        """
        Returns the total number of protocol bytes received from the server
        by this client.

        :return: The number of bytes consumed by protocol traffic read from
            the server during the lifetime of this client.
        :rtype: int
        """
        return self._protocol_bytes_received

    def get_dataset(self, dataset_id):
        """
        Returns the Dataset with the specified ID from the server.

        :param str dataset_id: The ID of the Dataset of interest.
        :return: The Dataset of interest.
        :rtype: :class:`ga4gh.protocol.Dataset`
        """
        return self._run_get_request(
            "datasets", protocol.Dataset, dataset_id)

    def get_reference_set(self, reference_set_id):
        """
        Returns the ReferenceSet with the specified ID from the server.

        :param str reference_set_id: The ID of the ReferenceSet of interest.
        :return: The ReferenceSet of interest.
        :rtype: :class:`ga4gh.protocol.ReferenceSet`
        """
        return self._run_get_request(
            "referencesets", protocol.ReferenceSet, reference_set_id)

    def get_reference(self, reference_id):
        """
        Returns the Reference with the specified ID from the server.

        :param str reference_id: The ID of the Reference of interest.
        :return: The Reference of interest.
        :rtype: :class:`ga4gh.protocol.Reference`
        """
        return self._run_get_request(
            "references", protocol.Reference, reference_id)

    def get_read_group_set(self, read_group_set_id):
        """
        Returns the ReadGroupSet with the specified ID from the server.

        :param str read_group_set_id: The ID of the ReadGroupSet of interest.
        :return: The ReadGroupSet of interest.
        :rtype: :class:`ga4gh.protocol.ReadGroupSet`
        """
        return self._run_get_request(
            "readgroupsets", protocol.ReadGroupSet, read_group_set_id)

    def get_read_group(self, read_group_id):
        """
        Returns the ReadGroup with the specified ID from the server.

        :param str read_group_id: The ID of the ReadGroup of interest.
        :return: The ReadGroup of interest.
        :rtype: :class:`ga4gh.protocol.ReadGroup`
        """
        return self._run_get_request(
            "readgroups", protocol.ReadGroup, read_group_id)

    def get_call_set(self, call_set_id):
        """
        Returns the CallSet with the specified ID from the server.

        :param str call_set_id: The ID of the CallSet of interest.
        :return: The CallSet of interest.
        :rtype: :class:`ga4gh.protocol.CallSet`
        """
        return self._run_get_request(
            "callsets", protocol.CallSet, call_set_id)

    def get_variant(self, variant_id):
        """
        Returns the Variant with the specified ID from the server.

        :param str variant_id: The ID of the Variant of interest.
        :return: The Variant of interest.
        :rtype: :class:`ga4gh.protocol.Variant`
        """
        return self._run_get_request(
            "variants", protocol.Variant, variant_id)

    def get_variant_set(self, variant_set_id):
        """
        Returns the VariantSet with the specified ID from the server.

        :param str variant_set_id: The ID of the VariantSet of interest.
        :return: The VariantSet of interest.
        :rtype: :class:`ga4gh.protocol.VariantSet`
        """
        return self._run_get_request(
            "variantsets", protocol.VariantSet, variant_set_id)

    def get_variant_annotation_set(self, variant_annotation_set_id):
        """
        Returns the VariantAnnotationSet with the specified ID from
        the server.

        :param str variant_annotation_set_id: The ID of the
            VariantAnnotationSet of interest.
        :return: The VariantAnnotationSet of interest.
        :rtype: :class:`ga4gh.protocol.VariantAnnotationSet`
        """
        return self._run_get_request(
            "variantannotationsets", protocol.VariantAnnotationSet,
            variant_annotation_set_id)

    def get_feature_set(self, feature_set_id):
        """
        Returns the FeatureSet with the specified ID from the server.

        :param str feature_set_id: The ID of the FeatureSet of interest.
        :return: The FeatureSet of interest.
        :rtype: :class:`ga4gh.protocol.FeatureSet`
        """
        return self._run_get_request(
            "featuresets", protocol.FeatureSet, feature_set_id)

    def get_feature(self, feature_id):
        """
        Returns the feature with the specified ID from the server.

        :param str feature_id: The ID of the requested feature
        :return: The requested ga4gh.protocol.Feature object.
        """
        return self._run_get_request(
            "features", protocol.Feature, feature_id)

    def get_rna_quantification_set(self, rna_quantification_set_id):
        """
        Returns the RnaQuantificationSet with the specified ID from the server.
        :param str rna_quantification_set_id: The ID of the
            RnaQuantificationSet of interest.
        :return: The RnaQuantificationSet of interest.
        :rtype: :class:`ga4gh.protocol.RnaQuantificationSet`
        """
        return self._run_get_request(
            "rnaquantificationsets", protocol.RnaQuantificationSet,
            rna_quantification_set_id)

    def get_rna_quantification(self, rna_quantification_id):
        """
        Returns the RnaQuantification with the specified ID from the server.
        :param str rna_quantification_id: The ID of the RnaQuantification of
            interest.
        :return: The RnaQuantification of interest.
        :rtype: :class:`ga4gh.protocol.RnaQuantification`
        """
        return self._run_get_request(
            "rnaquantifications", protocol.RnaQuantification,
            rna_quantification_id)

    def get_expression_level(self, expression_level_id):
        """
        Returns the ExpressionLevel with the specified ID from the server.
        :param str expression_level_id: The ID of the ExpressionLevel of
            interest.
        :return: The ExpressionLevel of interest.
        :rtype: :class:`ga4gh.protocol.ExpressionLevel`
        """
        return self._run_get_request(
            "expressionlevels", protocol.ExpressionLevel,
            expression_level_id)

    def search_variants(
            self, variant_set_id, start=None, end=None, reference_name=None,
            call_set_ids=None):
        """
        Returns an iterator over the Variants fulfilling the specified
        conditions from the specified VariantSet.

        :param str variant_set_id: The ID of the
            :class:`ga4gh.protocol.VariantSet` of interest.
        :param int start: Required. The beginning of the window (0-based,
            inclusive) for which overlapping variants should be returned.
            Genomic positions are non-negative integers less than reference
            length. Requests spanning the join of circular genomes are
            represented as two requests one on each side of the join
            (position 0).
        :param int end: Required. The end of the window (0-based, exclusive)
            for which overlapping variants should be returned.
        :param str reference_name: The name of the
            :class:`ga4gh.protocol.Reference` we wish to return variants from.
        :param list call_set_ids: Only return variant calls which belong to
            call sets with these IDs. If an empty array, returns variants
            without any call objects. If null, returns all variant calls.

        :return: An iterator over the :class:`ga4gh.protocol.Variant` objects
            defined by the query parameters.
        :rtype: iter
        """
        request = protocol.SearchVariantsRequest()
        request.reference_name = pb.string(reference_name)
        request.start = pb.int(start)
        request.end = pb.int(end)
        request.variant_set_id = variant_set_id
        request.call_set_ids.extend(pb.string(call_set_ids))
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "variants", protocol.SearchVariantsResponse)

    def search_variant_annotations(
            self, variant_annotation_set_id, reference_name="",
            reference_id="", start=0, end=0, effects=[]):
        """
        Returns an iterator over the Variant Annotations fulfilling
        the specified conditions from the specified VariantSet.

        :param str variant_annotation_set_id: The ID of the
            :class:`ga4gh.protocol.VariantAnnotationSet` of interest.
        :param int start: Required. The beginning of the window (0-based,
            inclusive) for which overlapping variants should be returned.
            Genomic positions are non-negative integers less than reference
            length. Requests spanning the join of circular genomes are
            represented as two requests one on each side of the join
            (position 0).
        :param int end: Required. The end of the window (0-based, exclusive)
            for which overlapping variants should be returned.
        :param str reference_name: The name of the
            :class:`ga4gh.protocol.Reference` we wish to return variants from.

        :return: An iterator over the
            :class:`ga4gh.protocol.VariantAnnotation` objects
            defined by the query parameters.
        :rtype: iter
        """
        request = protocol.SearchVariantAnnotationsRequest()
        request.variant_annotation_set_id = variant_annotation_set_id
        request.reference_name = reference_name
        request.reference_id = reference_id
        request.start = start
        request.end = end
        for effect in effects:
            request.effects.add().CopyFrom(protocol.OntologyTerm(**effect))
        for effect in request.effects:
            if not effect.id:
                raise exceptions.BadRequestException(
                    "Each ontology term should have an id set")
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "variantannotations",
            protocol.SearchVariantAnnotationsResponse)

    def search_features(
            self, feature_set_id=None, parent_id="", reference_name="",
            start=0, end=0, feature_types=[], name="", gene_symbol=""):
        """
        Returns the result of running a search_features method
        on a request with the passed-in parameters.

        :param str feature_set_id: ID of the feature Set being searched
        :param str parent_id: ID (optional) of the parent feature
        :param str reference_name: name of the reference to search
            (ex: "chr1")
        :param int start: search start position on reference
        :param int end: end position on reference
        :param feature_types: array of terms to limit search by (ex: "gene")
        :param str name: only return features with this name
        :param str gene_symbol: only return features on this gene
        :return: an iterator over Features as returned in the
            SearchFeaturesResponse object.
        """
        request = protocol.SearchFeaturesRequest()
        request.feature_set_id = feature_set_id
        request.parent_id = parent_id
        request.reference_name = reference_name
        request.name = name
        request.gene_symbol = gene_symbol
        request.start = start
        request.end = end
        request.feature_types.extend(feature_types)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "features",
            protocol.SearchFeaturesResponse)

    def search_datasets(self):
        """
        Returns an iterator over the Datasets on the server.

        :return: An iterator over the :class:`ga4gh.protocol.Dataset`
            objects on the server.
        """
        request = protocol.SearchDatasetsRequest()
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "datasets", protocol.SearchDatasetsResponse)

    def search_variant_sets(self, dataset_id):
        """
        Returns an iterator over the VariantSets fulfilling the specified
        conditions from the specified Dataset.

        :param str dataset_id: The ID of the :class:`ga4gh.protocol.Dataset`
            of interest.
        :return: An iterator over the :class:`ga4gh.protocol.VariantSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchVariantSetsRequest()
        request.dataset_id = dataset_id
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "variantsets", protocol.SearchVariantSetsResponse)

    def search_variant_annotation_sets(self, variant_set_id):
        """
        Returns an iterator over the Annotation Sets fulfilling the specified
        conditions from the specified variant set.

        :param str variant_set_id: The ID of the
            :class:`ga4gh.protocol.VariantSet` of interest.
        :return: An iterator over the :class:`ga4gh.protocol.AnnotationSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchVariantAnnotationSetsRequest()
        request.variant_set_id = variant_set_id
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "variantannotationsets",
            protocol.SearchVariantAnnotationSetsResponse)

    def search_feature_sets(self, dataset_id):
        """
        Returns an iterator over the FeatureSets fulfilling the specified
        conditions from the specified Dataset.

        :param str dataset_id: The ID of the
            :class:`ga4gh.protocol.Dataset` of interest.
        :return: An iterator over the :class:`ga4gh.protocol.FeatureSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchFeatureSetsRequest()
        request.dataset_id = dataset_id
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "featuresets", protocol.SearchFeatureSetsResponse)

    def search_reference_sets(
            self, accession=None, md5checksum=None, assembly_id=None):
        """
        Returns an iterator over the ReferenceSets fulfilling the specified
        conditions.

        :param str accession: If not null, return the reference sets for which
            the `accession` matches this string (case-sensitive, exact match).
        :param str md5checksum: If not null, return the reference sets for
            which the `md5checksum` matches this string (case-sensitive, exact
            match). See :class:`ga4gh.protocol.ReferenceSet::md5checksum` for
            details.
        :param str assembly_id: If not null, return the reference sets for
            which the `assembly_id` matches this string (case-sensitive,
            exact match).
        :return: An iterator over the :class:`ga4gh.protocol.ReferenceSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchReferenceSetsRequest()
        request.accession = pb.string(accession)
        request.md5checksum = pb.string(md5checksum)
        request.assembly_id = pb.string(assembly_id)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "referencesets", protocol.SearchReferenceSetsResponse)

    def search_references(
            self, reference_set_id, accession=None, md5checksum=None):
        """
        Returns an iterator over the References fulfilling the specified
        conditions from the specified Dataset.

        :param str reference_set_id: The ReferenceSet to search.
        :param str accession: If not None, return the references for which the
            `accession` matches this string (case-sensitive, exact match).
        :param str md5checksum: If not None, return the references for which
            the `md5checksum` matches this string (case-sensitive, exact
            match).
        :return: An iterator over the :class:`ga4gh.protocol.Reference`
            objects defined by the query parameters.
        """
        request = protocol.SearchReferencesRequest()
        request.reference_set_id = reference_set_id
        request.accession = pb.string(accession)
        request.md5checksum = pb.string(md5checksum)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "references", protocol.SearchReferencesResponse)

    def search_call_sets(self, variant_set_id, name=None, bio_sample_id=None):
        """
        Returns an iterator over the CallSets fulfilling the specified
        conditions from the specified VariantSet.

        :param str variant_set_id: Find callsets belonging to the
            provided variant set.
        :param str name: Only CallSets matching the specified name will
            be returned.
        :param str bio_sample_id: Only CallSets matching this id will
            be returned.
        :return: An iterator over the :class:`ga4gh.protocol.CallSet`
            objects defined by the query parameters.
        """
        request = protocol.SearchCallSetsRequest()
        request.variant_set_id = variant_set_id
        request.name = pb.string(name)
        request.bio_sample_id = pb.string(bio_sample_id)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "callsets", protocol.SearchCallSetsResponse)

    def search_bio_samples(self, dataset_id, name=None, individual_id=None):
        """
        Returns an iterator over the BioSamples fulfilling the specified
        conditions.

        :param str dataset_id: The dataset to search within.
        :param str name: Only BioSamples matching the specified name will
            be returned.
        :param str individual_id: Only BioSamples matching matching this
            id will be returned.
        :return: An iterator over the :class:`ga4gh.protocol.BioSample`
            objects defined by the query parameters.
        """
        request = protocol.SearchBioSamplesRequest()
        request.dataset_id = dataset_id
        request.name = pb.string(name)
        request.individual_id = pb.string(individual_id)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "biosamples", protocol.SearchBioSamplesResponse)

    def search_individuals(self, dataset_id, name=None):
        """
        Returns an iterator over the Individuals fulfilling the specified
        conditions.

        :param str dataset_id: The dataset to search within.
        :param str name: Only Individuals matching the specified name will
            be returned.
        :return: An iterator over the :class:`ga4gh.protocol.BioSample`
            objects defined by the query parameters.
        """
        request = protocol.SearchIndividualsRequest()
        request.dataset_id = dataset_id
        request.name = pb.string(name)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "individuals", protocol.SearchIndividualsResponse)

    def search_read_group_sets(
            self, dataset_id, name=None, bio_sample_id=None):
        """
        Returns an iterator over the ReadGroupSets fulfilling the specified
        conditions from the specified Dataset.

        :param str name: Only ReadGroupSets matching the specified name
            will be returned.
        :param str bio_sample_id: Only ReadGroups matching the specified
            bioSample will be included in the response.
        :return: An iterator over the :class:`ga4gh.protocol.ReadGroupSet`
            objects defined by the query parameters.
        :rtype: iter
        """
        request = protocol.SearchReadGroupSetsRequest()
        request.dataset_id = dataset_id
        request.name = pb.string(name)
        request.bio_sample_id = pb.string(bio_sample_id)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "readgroupsets", protocol.SearchReadGroupSetsResponse)

    def search_reads(
            self, read_group_ids, reference_id=None, start=None, end=None):
        """
        Returns an iterator over the Reads fulfilling the specified
        conditions from the specified read_group_ids.

        :param str read_group_ids: The IDs of the
            :class:`ga4gh.protocol.ReadGroup` of interest.
        :param str reference_id: The name of the
            :class:`ga4gh.protocol.Reference` we wish to return reads
            mapped to.
        :param int start: The start position (0-based) of this query. If a
            reference is specified, this defaults to 0. Genomic positions are
            non-negative integers less than reference length. Requests spanning
            the join of circular genomes are represented as two requests one on
            each side of the join (position 0).
        :param int end: The end position (0-based, exclusive) of this query.
            If a reference is specified, this defaults to the reference's
            length.
        :return: An iterator over the
            :class:`ga4gh.protocol.ReadAlignment` objects defined by
            the query parameters.
        :rtype: iter
        """
        request = protocol.SearchReadsRequest()
        request.read_group_ids.extend(read_group_ids)
        request.reference_id = pb.string(reference_id)
        request.start = pb.int(start)
        request.end = pb.int(end)
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "reads", protocol.SearchReadsResponse)

    def search_phenotype_association_sets(self, dataset_id):
        """
        Returns an iterator over the PhenotypeAssociationSets on the server.
        """
        request = protocol.SearchPhenotypeAssociationSetsRequest()
        request.dataset_id = dataset_id
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "phenotypeassociationsets",
            protocol.SearchPhenotypeAssociationSetsResponse)

    def search_genotype_phenotype(
            self, phenotype_association_set_id=None, feature_ids=None,
            phenotype_ids=None, evidence=None):
        """
        Returns an iterator over the GeneotypePhenotype associations from
        the server
        """
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = phenotype_association_set_id
        if feature_ids:
            request.feature_ids.extend(feature_ids)
        if phenotype_ids:
            request.phenotype_ids.extend(phenotype_ids)
        if evidence:
            request.evidence.extend(evidence)
        request.page_size = pb.int(self._page_size)
        self._logger.debug("search_genotype_phenotype {}".format(request))
        return self._run_search_request(
            request, "featurephenotypeassociations",
            protocol.SearchGenotypePhenotypeResponse)

    def search_phenotype(
            self, phenotype_association_set_id=None, phenotype_id=None,
            description=None, type_=None, age_of_onset=None):
        """
        Returns an iterator over the Phenotypes from the server
        """
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = phenotype_association_set_id
        if phenotype_id:
            request.id = phenotype_id
        if description:
            request.description = description
        if type_:
            request.type.mergeFrom(type_)
        if age_of_onset:
            request.age_of_onset = age_of_onset
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "phenotypes",
            protocol.SearchPhenotypesResponse)

    def search_rna_quantification_sets(self, dataset_id):
        """
        Returns an iterator over the RnaQuantificationSet objects from the
        server
        """
        request = protocol.SearchRnaQuantificationSetsRequest()
        request.dataset_id = dataset_id
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "rnaquantificationsets",
            protocol.SearchRnaQuantificationSetsResponse)

    def search_rna_quantifications(
            self, rna_quantification_set_id="", bio_sample_id=""):
        """
        Returns an iterator over the RnaQuantification objects from the server

        :param str rna_quantification_set_id: The ID of the
            :class:`ga4gh.protocol.RnaQuantificationSet` of interest.
        """
        request = protocol.SearchRnaQuantificationsRequest()
        request.rna_quantification_set_id = rna_quantification_set_id
        if bio_sample_id:
            request.bio_sample_id = bio_sample_id
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "rnaquantifications",
            protocol.SearchRnaQuantificationsResponse)

    def search_expression_levels(
            self, rna_quantification_id="", feature_ids=[], threshold=0.0):
        """
        Returns an iterator over the ExpressionLevel objects from the server

        :param str feature_ids: The IDs of the
            :class:`ga4gh.protocol.Feature` of interest.
        :param str rna_quantification_id: The ID of the
            :class:`ga4gh.protocol.RnaQuantification` of interest.
        :param float threshold: Minimum expression of responses to return.
        """
        request = protocol.SearchExpressionLevelsRequest()
        request.rna_quantification_id = rna_quantification_id
        request.feature_ids.extend(feature_ids)
        request.threshold = threshold
        request.page_size = pb.int(self._page_size)
        return self._run_search_request(
            request, "expressionlevels",
            protocol.SearchExpressionLevelsResponse)


class HttpClient(AbstractClient):
    """
    The GA4GH HTTP client. This class provides methods corresponding to the
    GA4GH search and object GET methods.

    .. todo:: Add a better description of the role of this class and include
        links to the high-level API documention.

    :param str urlPrefix: The base URL of the GA4GH server we wish to
        communicate with. This should include the 'http' or 'https' prefix.
    :param int logLevel: The amount of debugging information to log using
        the :mod:`logging` module. This is :data:`logging.WARNING` by default.
    :param str authentication_key: The authentication key provided by the
        server after logging in.
    """

    def __init__(
            self, url_prefix, logLevel=logging.WARNING,
            authentication_key=None):
        super(HttpClient, self).__init__(logLevel)
        self._url_prefix = url_prefix
        self._authentication_key = authentication_key
        self._session = requests.Session()
        self._setup_http_session()
        requests_log = logging.getLogger("requests.packages.urllib3")
        requests_log.setLevel(logLevel)
        requests_log.propagate = True

    def _setup_http_session(self):
        """
        Sets up the common HTTP session parameters used by requests.
        """
        headers = {"Content-type": "application/json"}
        self._session.headers.update(headers)
        # TODO is this unsafe????
        self._session.verify = False

    def _check_response_status(self, response):
        """
        Checks the speficied HTTP response from the requests package and
        raises an exception if a non-200 HTTP code was returned by the
        server.
        """
        if response.status_code != requests.codes.ok:
            self._logger.error("%s %s", response.status_code, response.text)
            raise exceptions.RequestNonSuccessException(
                "Url {0} had status_code {1}".format(
                    response.url, response.status_code))

    def _get_http_parameters(self):
        """
        Returns the basic HTTP parameters we need all requests.
        """
        return {'key': self._authentication_key}

    def _run_search_page_request(
            self, protocol_request, object_name, protocol_response_class):
        url = posixpath.join(self._url_prefix, object_name + '/search')
        data = protocol.toJson(protocol_request)
        self._logger.debug("request:{}".format(data))
        response = self._session.post(
            url, params=self._get_http_parameters(), data=data)
        self._check_response_status(response)
        return self._deserialize_response(
            response.text, protocol_response_class)

    def _run_get_request(self, object_name, protocol_response_class, id_):
        url_suffix = "{object_name}/{id}".format(
            object_name=object_name, id=id_)
        url = posixpath.join(self._url_prefix, url_suffix)
        response = self._session.get(url, params=self._get_http_parameters())
        self._check_response_status(response)
        return self._deserialize_response(
            response.text, protocol_response_class)

    def _run_list_reference_bases_page_request(self, request):
        url_suffix = "listreferencebases"
        url = posixpath.join(self._url_prefix, url_suffix)
        response = self._session.post(
            url, params=self._get_http_parameters(),
            data=protocol.toJson(request))
        self._check_response_status(response)
        return self._deserialize_response(
            response.text, protocol.ListReferenceBasesResponse)


class LocalClient(AbstractClient):

    def __init__(self, backend):
        super(LocalClient, self).__init__()
        self._backend = backend
        self._get_method_map = {
            "callsets": self._backend.runGetCallSet,
            "datasets": self._backend.runGetDataset,
            "referencesets": self._backend.runGetReferenceSet,
            "references": self._backend.runGetReference,
            "variantsets": self._backend.runGetVariantSet,
            "featuresets": self._backend.runGetFeatureSet,
            "variants": self._backend.runGetVariant,
            "features": self._backend.runGetFeature,
            "readgroupsets": self._backend.runGetReadGroupSet,
            "readgroups": self._backend.runGetReadGroup,
            "variantannotationsets": self._backend.runGetVariantAnnotationSet,
            "biosamples": self._backend.runGetBioSample,
            "individuals": self._backend.runGetIndividual,
            "rnaquantificationsets": self._backend.runGetRnaQuantificationSet,
            "rnaquantifications": self._backend.runGetRnaQuantification,
            "expressionlevels": self._backend.runGetExpressionLevel,
        }
        self._search_method_map = {
            "callsets": self._backend.runSearchCallSets,
            "datasets": self._backend.runSearchDatasets,
            "referencesets": self._backend.runSearchReferenceSets,
            "references": self._backend.runSearchReferences,
            "variantsets": self._backend.runSearchVariantSets,
            "featuresets": self._backend.runSearchFeatureSets,
            "variants": self._backend.runSearchVariants,
            "features": self._backend.runSearchFeatures,
            "readgroupsets": self._backend.runSearchReadGroupSets,
            "reads": self._backend.runSearchReads,
            "variantannotations": self._backend.runSearchVariantAnnotations,
            "variantannotationsets":
                self._backend.runSearchVariantAnnotationSets,
            "biosamples": self._backend.runSearchBioSamples,
            "individuals": self._backend.runSearchIndividuals,
            "featurephenotypeassociations":
                self._backend.runSearchGenotypePhenotypes,
            "phenotypes": self._backend.runSearchPhenotypes,
            "phenotypeassociationsets":
                self._backend.runSearchPhenotypeAssociationSets,
            "rnaquantificationsets":
                self._backend.runSearchRnaQuantificationSets,
            "rnaquantifications": self._backend.runSearchRnaQuantifications,
            "expressionlevels": self._backend.runSearchExpressionLevels,
        }

    def _run_get_request(self, object_name, protocol_response_class, id_):
        get_method = self._get_method_map[object_name]
        response_json = get_method(id_)
        return self._deserialize_response(
            response_json, protocol_response_class)

    def _run_search_page_request(
            self, protocol_request, object_name, protocol_response_class):
        search_method = self._search_method_map[object_name]
        response_json = search_method(protocol.toJson(protocol_request))
        return self._deserialize_response(
            response_json, protocol_response_class)

    def _run_list_reference_bases_page_request(self, request):
        response_json = self._backend.runListReferenceBases(
            protocol.toJson(request))
        return self._deserialize_response(
            response_json, protocol.ListReferenceBasesResponse)



#!/usr/bin/env python
"""
script to download gene level RNA quantifications from the ENCODE DCC site
https://www.encodeproject.org and convert to format used by the GA4GH API
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import os
import sys
import json
import errno
import urllib2
import optparse

import rnaseq2ga


class Color(object):
    """
    Color printing in terminal
    """

    RED = '\033[91m'
    BLUE = '\033[94m'
    GREEN = '\033[92m'
    NORMAL = '\033[0m'
    YELLOW = '\033[93m'

    @staticmethod
    def red(astr):
        return '{}{}{}'.format(Color.RED, str(astr), Color.NORMAL)

    @staticmethod
    def green(astr):
        return '{}{}{}'.format(Color.GREEN, str(astr), Color.NORMAL)

    @staticmethod
    def yellow(astr):
        return '{}{}{}'.format(Color.YELLOW, str(astr), Color.NORMAL)

    @staticmethod
    def blue(astr):
        return '{}{}{}'.format(Color.BLUE, str(astr), Color.NORMAL)


def log(message):
    print("== {}".format(Color.green(message)), file=sys.stderr)


def getFilesFromHost(data, host, outputType, request=False, subset=None):
    """
        analysisId is the ENCODE accession number for the quantification
        result file.  There may be multiple quantifications on a single ENCODE
        project or experiment page.
    """
    fileList = [f for f in data.get('files') if
                (f.get('output_type') == outputType)]
    if subset:
        fileList = fileList[:subset]
    for f in fileList:
        analysisId = f.get('accession')
        if not request:
            yield analysisId
        else:
            href = f.get('href')
            url = "{host}{file}".format(host=host, file=href)
            yield (analysisId, urllib2.urlopen(url))


def getDataFromHost(rnaDB, url, headers, host, outputType,
                    description, annotationId, subset=None,
                    readGroupId=""):
    req = urllib2.Request(url, headers=headers)
    try:
        response = urllib2.urlopen(req)
    except urllib2.URLError as e:
        if hasattr(e, 'reason'):
            print('We failed to reach a server.')
            print('Reason: {}'.format(e.reason))
        elif hasattr(e, 'code'):
            print("The server couldn't fulfill the request.")
            print('Error code: {}'.format(e.code))
    else:
        jsonData = json.load(response)
        writeRnaseqTables(rnaDB, getFilesFromHost(jsonData, host, outputType,
                          subset=subset), description, annotationId,
                          readGroupId=readGroupId)
        writer = rnaseq2ga.RsemWriter(rnaDB)
        writeGeneExpressionTables(writer, getFilesFromHost(jsonData, host,
                                  outputType, subset=subset, request=True))


def makeDir(path):
    try:
        os.makedirs(path)
    except OSError as exception:
        if exception.errno != errno.EEXIST:
            raise


def writeRnaQuant(rnaDB, analysisId, description, annotationId,
                  readGroupId="", programs=""):
    datafields = (analysisId, annotationId, description, analysisId,
                  readGroupId, programs)
    rnaDB.addRnaQuantification(datafields)


def writeRnaseqTables(rnaDB, analysisIds, description, annotationId,
                      readGroupId="", programs=""):
    log("Writing rnaseq tables")
    for analysisId in analysisIds:
        writeRnaQuant(rnaDB, analysisId, description, annotationId,
                      readGroupId=readGroupId, programs=programs)


def writeGeneExpressionTables(writer, data):
    log("Writing gene expression tables")
    for rnaQuantId, quantfile in data:
        print("processing {}".format(rnaQuantId))
        writer.writeExpression(rnaQuantId, quantfile)


def makeParser(usage):
    """
    Default values will download and parse 4 quantifications from the
    ENCODE Evaluation dataset as an example set of test data.
    """

    parser = optparse.OptionParser(usage=usage)
    parser.add_option("--dataset", dest="dataset")
    parser.add_option("--description", dest="description")
    parser.add_option("--annotationId", dest="annotationId")
    parser.add_option("--file_limit", type="int", dest="subset")
    parser.add_option("--readGroupId", dest="readGroupId")
    parser.add_option("--expressionType", dest="expressionType")
    parser.set_defaults(dataset="ENCSR000AJW", annotationId="Gencodev16",
                        description="RNAseq data from ENCODE evaluation",
                        subset=4, readGroupId="", expressionType="gene")
    return parser


def main(argv):

    usage = "Usage: {} <db-file>".format(argv[0])
    if len(argv) < 2:
        print(Color.red(usage))
        sys.exit(1)
    parser = makeParser(usage)
    (options, args) = parser.parse_args(argv[1:])
    host = "https://www.encodeproject.org"
    dataset = options.dataset
    url = "{host}/datasets/{dataset}/?frame=embedded".format(host=host,
                                                             dataset=dataset)
    headers = {
        'Accept': 'application/json; charset=utf-8'
    }
    dataType = ""
    if options.expressionType == "gene":
        dataType = "gene quantifications"
    sqlFilename = argv[1]
    rnaDB = rnaseq2ga.RnaSqliteStore(sqlFilename)
    subset = options.subset
    log("Downloading GA4GH test dataset - RNA Quantification API")
    print("ENCODE dataset: {}".format(Color.blue(dataset)))
    print("data type:      {}".format(Color.blue(dataType)))
    print("subset size:    {}".format(Color.blue(subset)))
    print("readGroupId  :  {}".format(Color.blue(options.readGroupId)))
    print("output db:  {}".format(Color.blue(sqlFilename)))
    # TODO: annotation name can be found at DCC...
    getDataFromHost(rnaDB, url, headers, host, dataType,
                    options.description, options.annotationId, subset,
                    options.readGroupId)
    rnaDB.batchAddRnaQuantification()
    log("DONE")


if __name__ == '__main__':
    main(sys.argv)



"""
configtest cli
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import unittest

import ga4gh.cli as cli
import ga4gh.configtest as configtest


class SimplerResult(unittest.TestResult):
    """
    The TestResult class gives formatted tracebacks as error messages, which
    is not what we want. Instead we just want the error message from the
    err praram. Hence this subclass.
    """
    def addError(self, test, err):
        self.errors.append((test,
                            "{0}: {1}".format(err[0].__name__, err[1])))

    def addFailure(self, test, err):
        self.failures.append((test,
                              "{0}: {1}".format(err[0].__name__, err[1])))


def configtest_main(parser=None):
    if parser is None:
        parser = cli.createArgumentParser(
            "GA4GH server configuration validator")
    parser.add_argument(
        "--config", "-c", default='DevelopmentConfig', type=str,
        help="The configuration to use")
    parser.add_argument(
        "--config-file", "-f", type=str, default=None,
        help="The configuration file to use")
    cli.addVersionArgument(parser)

    args = parser.parse_args()
    configStr = 'ga4gh.serverconfig:{0}'.format(args.config)

    configtest.TestConfig.configStr = configStr
    configtest.TestConfig.configFile = args.config_file
    configtest.TestConfig.configEnv = "GA4GH_CONFIGURATION"

    loader = unittest.TestLoader()
    tests = loader.loadTestsFromModule(configtest)
    results = SimplerResult()
    tests.run(results)

    logging.basicConfig(level=logging.INFO)
    log = logging.getLogger(__name__)
    log.info('{0} Tests run. {1} errors, {2} failures, {3} skipped'.
             format(results.testsRun,
                    len(results.errors),
                    len(results.failures),
                    len(results.skipped)))
    for result in results.errors:
        if result is not None:
            log.critical('Error: {0}: {1}'.format(result[0].id(), result[1]))
    for result in results.failures:
        if result is not None:
            log.critical('Failure: {0}: {1}'.format(result[0].id(), result[1]))
    for result in results.skipped:
        if result is not None:
            log.info('Skipped: {0}: {1}'.format(result[0].id(), result[1]))



"""
Data-driven tests for sequence annotation Features.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import ga4gh.datarepo as datarepo
import ga4gh.datamodel as datamodel
import ga4gh.datamodel.datasets as datasets
import ga4gh.datamodel.references as references
import ga4gh.datamodel.sequence_annotations as sequence_annotations
import ga4gh.protocol as protocol
import tests.datadriven as datadriven
import tests.paths as paths

_datasetName = "ds"

_discontinuousTestData = {
    "featureSetName": "discontinuous",
    "referenceName": "apidb|Pf3D7_13",
    "totalFeatures": 30,
    "sampleFeatureId": 4409955920,
    "sampleParentId": 4409956112,
    "sampleStart": 820942,
    "sampleEnd": 821379,
    "sampleStrand": protocol.POS_STRAND,
    "sampleSiblings": 2,
    "region": [0, 2**32],
    "ontologyRestriction": ["gene", ],
    "featuresWithOntology": 3
}

_gencodeV21Set1TestData = {
    "featureSetName": "gencodeV21Set1",
    "referenceName": "chr1",
    "totalFeatures": 543,
    "sampleFeatureId": 4397111632,
    "sampleParentId": 4397111312,
    "sampleStart": 804776,
    "sampleEnd": 804832,
    "sampleStrand": protocol.POS_STRAND,
    "sampleSiblings": 5,
    "region": [0, 2**32],
    "ontologyRestriction": ["gene", ],
    "featuresWithOntology": 58
}

_sacCerTestTestData = {
    "featureSetName": "sacCerTest",
    "referenceName": "chrI",
    "totalFeatures": 33,
    "sampleFeatureId": 4324354512,
    "sampleParentId": None,
    "sampleStart": 337,
    "sampleEnd": 801,
    "sampleStrand": protocol.NEG_STRAND,
    "sampleSiblings": 33,
    "region": [0, 2**32],
    "ontologyRestriction": ["gene", ],
    "featuresWithOntology": 11
}

_specialCasesTestTestData = {
    "featureSetName": "specialCasesTest",
    "referenceName": "2L",
    "totalFeatures": 4,
    "sampleFeatureId": 4554844304,
    "sampleParentId": None,
    "sampleStart": 22229583,
    "sampleEnd": 22229699,
    "sampleStrand": protocol.NEG_STRAND,
    "sampleSiblings": 4,
    "region": [0, 2**32],
    "ontologyRestriction": ["gene", ],
    "featuresWithOntology": 0
}

_testDataForFeatureSetName = {
    "discontinuous": _discontinuousTestData,
    "gencodeV21Set1": _gencodeV21Set1TestData,
    "sacCerTest": _sacCerTestTestData,
    "specialCasesTest": _specialCasesTestTestData
}


def _getFeatureCompoundId(dataSetName, featureSetName, featureId):
    splits = [dataSetName, featureSetName, str(featureId)]
    joined = datamodel.CompoundId.join(splits)
    obfuscated = datamodel.CompoundId.obfuscate(joined)
    return obfuscated


def testFeatureSets():
    testDataDir = "tests/data/datasets/dataset1/sequenceAnnotations"
    for test in datadriven.makeTests(testDataDir, FeatureSetTests, "*.db"):
        yield test


class FeatureSetTests(datadriven.DataDrivenTest):
    """
    Re-parses source GFF3 files, compares the results with the contents
    of corresponding sequence_annotations.Feature objects.
    """
    def __init__(self, featureSetLocalName, dataPath):
        """
        :param localId: Name of the GFF3 resource corresponding to a pair
        of files, .db and .gff3
        :param dataPath: string representing full path to the .db file
        :return:
        """
        self._dataset = datasets.Dataset(_datasetName)
        self._repo = datarepo.SqlDataRepository(paths.testDataRepo)
        self._repo.open(datarepo.MODE_READ)
        self._ontology = self._repo.getOntologyByName(paths.ontologyName)
        self._referenceSet = references.AbstractReferenceSet("test_rs")
        featureSetLocalName = featureSetLocalName[:-3]  # remove '.db'
        self._testData = _testDataForFeatureSetName[featureSetLocalName]
        super(FeatureSetTests, self).__init__(featureSetLocalName, dataPath)

    def getProtocolClass(self):
        return protocol.FeatureSet

    def getDataModelInstance(self, localId, dataPath):
        featureSet = sequence_annotations.Gff3DbFeatureSet(
            self._dataset, localId)
        featureSet.setOntology(self._ontology)
        featureSet.setReferenceSet(self._referenceSet)
        featureSet.populateFromFile(dataPath)
        return featureSet

    def testGetFeatureById(self):
        idString = _getFeatureCompoundId(
            _datasetName,
            self._testData["featureSetName"],
            self._testData["sampleFeatureId"])
        compoundId = datamodel.FeatureCompoundId.parse(idString)
        feature = self._gaObject.getFeature(compoundId)
        self.assertIsNotNone(feature)
        self.assertEqual(feature.id, idString)
        if self._testData["sampleParentId"] is not None:
            self.assertEqual(
                datamodel.FeatureCompoundId.parse(
                    feature.parent_id).featureId,
                str(self._testData["sampleParentId"]))
        else:
            self.assertEqual(feature.parent_id, '')
        self.assertEqual(
            feature.reference_name,
            self._testData["referenceName"])
        self.assertEqual(
            feature.start,
            self._testData["sampleStart"])
        self.assertEqual(
            feature.end,
            self._testData["sampleEnd"])
        self.assertEqual(
            feature.strand,
            self._testData["sampleStrand"])

    def testFetchAllFeaturesInRegion(self):
        features = []
        for feature in self._gaObject.getFeatures(
                self._testData["referenceName"],
                self._testData["region"][0],
                self._testData["region"][1],
                None, 1000):
            features.append(feature)
        self.assertEqual(len(features), self._testData["totalFeatures"])

    def testFetchFeaturesRestrictedByOntology(self):
        features = []
        for feature in self._gaObject.getFeatures(
                self._testData["referenceName"],
                self._testData["region"][0],
                self._testData["region"][1],
                None, 1000,
                featureTypes=self._testData["ontologyRestriction"]):
            features.append(feature)
        self.assertEqual(len(features),
                         self._testData["featuresWithOntology"])

    def testFetchFeaturesRestrictedByParent(self):
        parentId = ""
        if self._testData["sampleParentId"] is not None:
            parentIdString = _getFeatureCompoundId(
                _datasetName,
                self._testData["featureSetName"],
                self._testData["sampleParentId"])
            parentId = datamodel.FeatureCompoundId.parse(
                parentIdString).featureId
        features = []
        for feature in self._gaObject.getFeatures(
                self._testData["referenceName"],
                self._testData["region"][0],
                self._testData["region"][1],
                None, 1000,
                parentId=parentId):
            features.append(feature)
        self.assertEqual(len(features),
                         self._testData["sampleSiblings"])



"""
Tests the cli
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import mock
import unittest
import shlex

import ga4gh.cli.server as cli_server
import ga4gh.cli.client as cli_client
import ga4gh.cli.repomanager as cli_repomanager
import ga4gh.cli.ga2vcf as cli_ga2vcf
import ga4gh.cli.ga2sam as cli_ga2sam
import ga4gh.protocol as protocol
import google.protobuf.descriptor as descriptor
import google.protobuf.internal.python_message as python_message
import tests.utils as utils


class TestServerArguments(unittest.TestCase):
    """
    Tests that the server can parse expected arguments
    """
    def testParseArguments(self):
        cliInput = """--port 7777 --host 123.4.5.6 --config MockConfigName
        --config-file /path/to/config --tls --dont-use-reloader"""
        parser = cli_server.getServerParser()
        args = parser.parse_args(cliInput.split())
        self.assertEqual(args.port, 7777)
        self.assertEqual(args.host, "123.4.5.6")
        self.assertEqual(args.config, "MockConfigName")
        self.assertEqual(args.config_file, "/path/to/config")
        self.assertTrue(args.tls)
        self.assertTrue(args.dont_use_reloader)


class TestGa2VcfArguments(unittest.TestCase):
    """
    Tests the ga2vcf cli can parse all arguments it is supposed to
    """
    def testParseArguments(self):
        cliInput = """--key KEY -O vcf --outputFile /dev/null
        --referenceName REFERENCENAME --callSetIds CALL,SET,IDS --start 0
        --end 1 --pageSize 2 BASEURL VARIANTSETID"""
        parser = cli_ga2vcf.getGa2VcfParser()
        args = parser.parse_args(cliInput.split())
        self.assertEqual(args.key, "KEY")
        self.assertEqual(args.outputFormat, "vcf")
        self.assertEqual(args.outputFile, "/dev/null")
        self.assertEqual(args.referenceName, "REFERENCENAME")
        self.assertEqual(args.callSetIds, "CALL,SET,IDS")
        self.assertEqual(args.start, 0)
        self.assertEqual(args.end, 1)
        self.assertEqual(args.pageSize, 2)
        self.assertEquals(args.baseUrl, "BASEURL")
        self.assertEquals(args.variantSetId, "VARIANTSETID")


class TestGa2SamArguments(unittest.TestCase):
    """
    Tests the ga2sam cli can parse all arguments it is supposed to
    """
    def testParseArguments(self):
        cliInput = """--key KEY --outputFormat sam
        --pageSize 1 --start 2 --end 3 --outputFile OUT.SAM
        --referenceId REFERENCEID BASEURL READGROUPID"""
        parser = cli_ga2sam.getGa2SamParser()
        args = parser.parse_args(cliInput.split())
        self.assertEqual(args.key, "KEY")
        self.assertEqual(args.outputFormat, "sam")
        self.assertEqual(args.outputFile, "OUT.SAM")
        self.assertEqual(args.referenceId, "REFERENCEID")
        self.assertEqual(args.start, 2)
        self.assertEqual(args.end, 3)
        self.assertEqual(args.pageSize, 1)
        self.assertEquals(args.baseUrl, "BASEURL")
        self.assertEquals(args.readGroupId, "READGROUPID")


class TestClientArguments(unittest.TestCase):
    """
    Tests the client cli can parse all arguments it is supposed to
    and can initialize the runner in preparation for a request
    """
    class ParseFailureException(Exception):
        pass

    def _raiseParseFailureException(self, exitVal):
        raise self.ParseFailureException(exitVal)

    def setUp(self):
        self.parser = cli_client.getClientParser()

    def testParseFailure(self):
        cliInput = "invalidCommand"
        with utils.suppressOutput():
            with mock.patch('sys.exit', self._raiseParseFailureException):
                with self.assertRaises(self.ParseFailureException):
                    self.parser.parse_args(cliInput.split())

    def testOutputFormat(self):
        # Most of the commands support the output format option.
        cliInput = "variants-search BASEURL --outputFormat=json"
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.outputFormat, "json")
        cliInput = "variants-search BASEURL -O text"
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.outputFormat, "text")

    def testVariantsSearchArguments(self):
        cliInput = (
            "variants-search --referenceName REFERENCENAME "
            "--callSetIds CALL,SET,IDS --start 0 "
            "--end 1 --pageSize 2 --variantSetId VARIANTSETID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.start, 0)
        self.assertEqual(args.end, 1)
        self.assertEqual(args.referenceName, "REFERENCENAME")
        self.assertEqual(args.callSetIds, "CALL,SET,IDS")
        self.assertEqual(args.pageSize, 2)
        self.assertEqual(args.variantSetId, "VARIANTSETID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.runner, cli_client.SearchVariantsRunner)

    def testVariantSetsSearchArguments(self):
        cliInput = (
            "variantsets-search --pageSize 1 --datasetId DATASETID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchVariantSetsRunner)

    def testReferenceSetsSearchArguments(self):
        cliInput = (
            "referencesets-search --pageSize 1 --accession ACCESSION "
            "--md5checksum MD5CHECKSUM --assemblyId ASSEMBLYID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.md5checksum, "MD5CHECKSUM")
        self.assertEqual(args.assemblyId, "ASSEMBLYID")
        self.assertEqual(args.accession, "ACCESSION")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchReferenceSetsRunner)

    def testReferencesSearchArguments(self):
        cliInput = (
            "references-search --pageSize 10 --accession ACCESSION "
            "--md5checksum MD5CHECKSUM BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 10)
        self.assertEqual(args.md5checksum, "MD5CHECKSUM")
        self.assertEqual(args.accession, "ACCESSION")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchReferencesRunner)

    def testReadGroupSetsSearchArguments(self):
        cliInput = (
            "readgroupsets-search --pageSize 1 --datasetId DATASETID "
            "--name NAME BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.name, "NAME")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchReadGroupSetsRunner)

    def testCallSetsSearchArguments(self):
        cliInput = (
            "callsets-search --pageSize 1 --name NAME "
            "--variantSetId VARIANTSETID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.variantSetId, "VARIANTSETID")
        self.assertEqual(args.name, "NAME")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchCallSetsRunner)

    def testReadsSearchArguments(self):
        cliInput = (
            "reads-search --pageSize 2 --start 5 --end 10 "
            "--readGroupIds READ,GROUP,IDS --referenceId REFERENCEID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 2)
        self.assertEqual(args.start, 5)
        self.assertEqual(args.end, 10)
        self.assertEqual(args.readGroupIds, "READ,GROUP,IDS")
        self.assertEqual(args.referenceId, "REFERENCEID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchReadsRunner)

    def testBioSamplesSearchArguments(self):
        cliInput = (
            "biosamples-search --pageSize 2 --name BIOSAMPLENAME "
            "--datasetId DATASETID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 2)
        self.assertEqual(args.name, "BIOSAMPLENAME")
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchBioSamplesRunner)

    def testIndividualsSearchArguments(self):
        cliInput = (
            "individuals-search --pageSize 2 --name INDIVIDUALNAME "
            "--datasetId DATASETID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 2)
        self.assertEqual(args.name, "INDIVIDUALNAME")
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchIndividualsRunner)

    def testDatasetsSearchArguments(self):
        cliInput = "datasets-search BASEURL"
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchDatasetsRunner)

    def testGenotypePhenotypeSearchArguments(self):
        cliInput = (
            "genotypephenotype-search --phenotype_association_set_id SET_ID "
            "--feature_ids A,B,C --phenotype_ids D,E,F --pageSize 1 -E E1 "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.phenotype_association_set_id, "SET_ID")
        self.assertEqual(args.feature_ids, "A,B,C")
        self.assertEqual(args.phenotype_ids, "D,E,F")
        self.assertEqual(args.evidence, "E1")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(
            args.runner, cli_client.SearchGenotypePhenotypeRunner)

    def testPhenotypeSearchArguments(self):
        cliInput = (
            "phenotype-search --phenotype_association_set_id SET_ID "
            "--phenotype_id ID1 --description FOO --type T --age_of_onset 2 "
            "--pageSize 1 BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.phenotype_association_set_id, "SET_ID")
        self.assertEqual(args.phenotype_id, "ID1")
        self.assertEqual(args.description, "FOO")
        self.assertEqual(args.type, "T")
        self.assertEqual(args.age_of_onset, "2")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(args.runner, cli_client.SearchPhenotypeRunner)

    def testPhenotypeAssociationSetsSearchArguments(self):
        cliInput = (
            "phenotypeassociationsets-search --datasetId SET_ID "
            "--pageSize 1 BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 1)
        self.assertEqual(args.datasetId, "SET_ID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(
            args.runner, cli_client.SearchPhenotypeAssociationSetsRunner)

    def verifyGetArguments(self, command, runnerClass):
        cliInput = "{} BASEURL ID".format(command)
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.id, "ID")
        self.assertEquals(args.runner, runnerClass)

    def testReferenceSetGetArguments(self):
        self.verifyGetArguments(
            "referencesets-get", cli_client.GetReferenceSetRunner)

    def testBioSamplesGetArguments(self):
        self.verifyGetArguments(
            "biosamples-get", cli_client.GetBioSampleRunner)

    def testIndividualsGetArguments(self):
        self.verifyGetArguments(
            "individuals-get", cli_client.GetIndividualRunner)

    def testReferenceGetArguments(self):
        self.verifyGetArguments(
            "references-get", cli_client.GetReferenceRunner)

    def testReadGroupSetGetArguments(self):
        self.verifyGetArguments(
            "readgroupsets-get", cli_client.GetReadGroupSetRunner)

    def testReadGroupGetArguments(self):
        self.verifyGetArguments(
            "readgroups-get", cli_client.GetReadGroupRunner)

    def testCallSetGetArguments(self):
        self.verifyGetArguments(
            "callsets-get", cli_client.GetCallSetRunner)

    def testDatasetsGetArguments(self):
        self.verifyGetArguments(
            "datasets-get", cli_client.GetDatasetRunner)

    def testVariantGetArguments(self):
        self.verifyGetArguments(
            "variants-get", cli_client.GetVariantRunner)

    def testRnaQuantificationGetArguments(self):
        self.verifyGetArguments(
            "rnaquantifications-get", cli_client.GetRnaQuantificationRunner)

    def testVariantSetsGet(self):
        self.verifyGetArguments(
            "variantsets-get", cli_client.GetVariantSetRunner)

    def testFeaturesGet(self):
        self.verifyGetArguments(
            "features-get", cli_client.GetFeatureRunner)

    def testFeatureSetsGet(self):
        self.verifyGetArguments(
            "featuresets-get", cli_client.GetFeatureSetRunner)

    def testExpressionLevelsGet(self):
        self.verifyGetArguments(
            "expressionlevels-get", cli_client.GetExpressionLevelRunner)

    def testRnaQuantificationSetsGet(self):
        self.verifyGetArguments(
            "rnaquantificationsets-get",
            cli_client.GetRnaQuantificationSetRunner)

    def testReferenceBasesListArguments(self):
        cliInput = (
            "references-list-bases BASEURL ID --start 1 --end 2 "
            "--outputFormat fasta")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.id, "ID")
        self.assertEqual(args.start, 1)
        self.assertEqual(args.end, 2)
        self.assertEquals(args.outputFormat, "fasta")
        self.assertEquals(args.runner, cli_client.ListReferenceBasesRunner)

    def testVariantAnnotationsSearch(self):
        cliInput = (
            "variantannotations-search "
            "--variantAnnotationSetId VARIANTANNOTATIONSETID "
            "--referenceName REFERENCENAME --start 1 "
            "--end 2 --effects EFFECTS "
            "--pageSize 3 BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(
            args.variantAnnotationSetId, "VARIANTANNOTATIONSETID")
        self.assertEqual(args.referenceName, "REFERENCENAME")
        self.assertEqual(args.start, 1)
        self.assertEqual(args.end, 2)
        self.assertEqual(args.effects, "EFFECTS")
        self.assertEqual(args.pageSize, 3)
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEquals(
            args.runner, cli_client.SearchVariantAnnotationsRunner)

    def testVariationAnnotationSetsSearch(self):
        cliInput = (
            "variantannotationsets-search "
            "--pageSize 3 BASEURL VARIANTSETID")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 3)
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.variantSetId, "VARIANTSETID")
        self.assertEquals(
            args.runner, cli_client.SearchVariantAnnotationSetsRunner)

    def testVariationAnnotationSetsGet(self):
        cliInput = (
            "variantannotationsets-get "
            "BASEURL VARIANTANNOTATIONSETID")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.id, "VARIANTANNOTATIONSETID")
        self.assertEquals(
            args.runner, cli_client.GetVariantAnnotationSetRunner)

    def testRnaQuantificationSearchArguments(self):
        cliInput = (
            "rnaquantifications-search --rnaQuantificationSetId ID BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.rnaQuantificationSetId, "ID")
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(
            args.runner, cli_client.SearchRnaQuantificationsRunner)

    def testExpressionLevelSearchArguments(self):
        cliInput = (
            "expressionlevels-search --rnaQuantificationId rID "
            "--threshold 0.0 BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.rnaQuantificationId, "rID")
        self.assertEqual(args.threshold, 0.0)
        self.assertEqual(args.baseUrl, "BASEURL")
        self.assertEqual(args.runner, cli_client.SearchExpressionLevelsRunner)

    def testFeaturesSearch(self):
        cliInput = (
            "features-search "
            "--pageSize 3 "
            "--featureSetId FEATURESETID "
            "--start 1 "
            "--end 2 "
            "--parentId PARENTID "
            "--featureTypes FEATURE,TYPES "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 3)
        self.assertEqual(args.featureSetId, "FEATURESETID")
        self.assertEqual(args.start, 1)
        self.assertEqual(args.end, 2)
        self.assertEqual(args.parentId, "PARENTID")
        self.assertEqual(args.featureTypes, "FEATURE,TYPES")
        self.assertEqual(args.baseUrl, "BASEURL")

    def testFeatureSetsSearch(self):
        cliInput = (
            "featuresets-search "
            "--pageSize 3 "
            "--datasetId DATASETID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 3)
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.baseUrl, "BASEURL")

    def testRnaQuantificationSetsSearch(self):
        cliInput = (
            "rnaquantificationsets-search "
            "--pageSize 3 "
            "--datasetId DATASETID "
            "BASEURL")
        args = self.parser.parse_args(cliInput.split())
        self.assertEqual(args.pageSize, 3)
        self.assertEqual(args.datasetId, "DATASETID")
        self.assertEqual(args.baseUrl, "BASEURL")


class TestRepoManagerCli(unittest.TestCase):

    def setUp(self):
        self.parser = cli_repomanager.RepoManager.getParser()
        self.registryPath = 'a/repo/path'
        self.datasetName = "datasetName"
        self.filePath = 'a/file/path'
        self.dirPath = 'a/dir/path/'
        self.individualName = "test"
        self.bioSampleName = "test"
        self.individual = protocol.toJson(
            protocol.Individual(
                name="test",
                created="2016-05-19T21:00:19Z",
                updated="2016-05-19T21:00:19Z"))
        self.bioSample = protocol.toJson(
            protocol.BioSample(
                name="test",
                created="2016-05-19T21:00:19Z",
                updated="2016-05-19T21:00:19Z"))

    def testInit(self):
        cliInput = "init {}".format(self.registryPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.runner, "init")

    def testVerify(self):
        cliInput = "verify {}".format(self.registryPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.runner, "verify")

    def testList(self):
        cliInput = "list {}".format(self.registryPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.runner, "list")

    def testAddDataset(self):
        cliInput = "add-dataset {} {}".format(
            self.registryPath, self.datasetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.runner, "addDataset")

    def testRemoveDataset(self):
        cliInput = "remove-dataset {} {} -f".format(
            self.registryPath, self.datasetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.runner, "removeDataset")
        self.assertEquals(args.force, True)

    def testAddReferenceSet(self):
        description = "description"
        cliInput = (
            "add-referenceset {} {} --description={} "
            "--ncbiTaxonId NCBITAXONID "
            "--isDerived True "
            "--assemblyId ASSEMBLYID "
            "--sourceAccessions SOURCEACCESSIONS "
            "--sourceUri SOURCEURI ").format(
            self.registryPath, self.filePath, description)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.filePath, self.filePath)
        self.assertEquals(args.description, description)
        self.assertEquals(args.ncbiTaxonId, "NCBITAXONID")
        self.assertEquals(args.isDerived, True)
        self.assertEquals(args.assemblyId, "ASSEMBLYID")
        self.assertEquals(args.sourceAccessions, "SOURCEACCESSIONS")
        self.assertEquals(args.sourceUri, "SOURCEURI")
        self.assertEquals(args.runner, "addReferenceSet")

    def testRemoveReferenceSet(self):
        referenceSetName = "referenceSetName"
        cliInput = "remove-referenceset {} {} -f".format(
            self.registryPath, referenceSetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.referenceSetName, referenceSetName)
        self.assertEquals(args.runner, "removeReferenceSet")
        self.assertEquals(args.force, True)

    def testAddReadGroupSet(self):
        cliInput = "add-readgroupset {} {} {} ".format(
            self.registryPath, self.datasetName, self.filePath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFile, self.filePath)
        self.assertEquals(args.indexFile, None)
        self.assertEquals(args.runner, "addReadGroupSet")

    def testAddReadGroupSetWithIndexFile(self):
        indexPath = self.filePath + ".bai"
        cliInput = "add-readgroupset {} {} {} -I {}".format(
            self.registryPath, self.datasetName, self.filePath,
            indexPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFile, self.filePath)
        self.assertEquals(args.indexFile, indexPath)
        self.assertEquals(args.runner, "addReadGroupSet")

    def testRemoveReadGroupSet(self):
        readGroupSetName = "readGroupSetName"
        cliInput = "remove-readgroupset {} {} {} -f".format(
            self.registryPath, self.datasetName, readGroupSetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.readGroupSetName, readGroupSetName)
        self.assertEquals(args.runner, "removeReadGroupSet")
        self.assertEquals(args.force, True)

    def testAddVariantSet(self):
        cliInput = "add-variantset {} {} {} ".format(
            self.registryPath, self.datasetName, self.filePath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFiles, [self.filePath])
        self.assertEquals(args.indexFiles, None)
        self.assertEquals(args.runner, "addVariantSet")

    def testAddVariantSetWithIndexFiles(self):
        file1 = "file1"
        file2 = "file2"
        indexFile1 = file1 + ".tbi"
        indexFile2 = file2 + ".tbi"
        cliInput = "add-variantset {} {} {} {} -I {} {}".format(
            self.registryPath, self.datasetName, file1, file2,
            indexFile1, indexFile2)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dataFiles, [file1, file2])
        self.assertEquals(args.indexFiles, [indexFile1, indexFile2])
        self.assertEquals(args.runner, "addVariantSet")

    def testRemoveVariantSet(self):
        variantSetName = "variantSetName"
        cliInput = "remove-variantset {} {} {}".format(
            self.registryPath, self.datasetName, variantSetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.variantSetName, variantSetName)
        self.assertEquals(args.runner, "removeVariantSet")
        self.assertEquals(args.force, False)

    def testAddOntology(self):
        cliInput = "add-ontology {} {}".format(
            self.registryPath, self.filePath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.filePath, self.filePath)
        self.assertEquals(args.runner, "addOntology")

    def testRemoveOntology(self):
        ontologyName = "the_ontology_name"
        cliInput = "remove-ontology {} {}".format(
            self.registryPath, ontologyName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.ontologyName, ontologyName)
        self.assertEquals(args.runner, "removeOntology")
        self.assertEquals(args.force, False)

    def testAddBioSample(self):
        cliInput = "add-biosample {} {} {} '{}'".format(
            self.registryPath,
            self.datasetName,
            self.bioSampleName,
            self.bioSample)
        args = self.parser.parse_args(shlex.split(cliInput))
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.bioSampleName, self.bioSampleName)
        self.assertEquals(args.bioSample, self.bioSample)
        self.assertEquals(args.runner, "addBioSample")

    def testRemoveBioSample(self):
        cliInput = "remove-biosample {} {} {}".format(
            self.registryPath,
            self.datasetName,
            self.bioSampleName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.bioSampleName, self.bioSampleName)
        self.assertEquals(args.runner, "removeBioSample")
        self.assertEquals(args.force, False)

    def testAddIndividual(self):
        cliInput = "add-individual {} {} {} '{}'".format(
            self.registryPath,
            self.datasetName,
            self.individualName,
            self.individual)
        args = self.parser.parse_args(shlex.split(cliInput))
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.individualName, self.individualName)
        self.assertEquals(args.individual, self.individual)
        self.assertEquals(args.runner, "addIndividual")

    def testRemoveIndividual(self):
        cliInput = "remove-individual {} {} {}".format(
            self.registryPath,
            self.datasetName,
            self.individualName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.individualName, self.individualName)
        self.assertEquals(args.runner, "removeIndividual")
        self.assertEquals(args.force, False)

    def testAddPhenotypeAssociationSet(self):
        cliInput = "add-phenotypeassociationset {} {} {} -n NAME".format(
            self.registryPath,
            self.datasetName,
            self.dirPath)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.dirPath, self.dirPath)
        self.assertEquals(args.name, "NAME")

    def testRemovePhenotypeAssociationSet(self):
        cliInput = "remove-phenotypeassociationset {} {} NAME".format(
            self.registryPath, self.datasetName)
        args = self.parser.parse_args(cliInput.split())
        self.assertEquals(args.registryPath, self.registryPath)
        self.assertEquals(args.datasetName, self.datasetName)
        self.assertEquals(args.name, "NAME")


class TestOutputFormats(unittest.TestCase):
    """
    Tests the different output formats of the cli
    """
    class FakeArgs(object):
        def __init__(self, outputFormat='text'):
            self.outputFormat = outputFormat
            self.id = 'id'
            self.key = 'key'
            self.baseUrl = 'baseUrl'
            self.verbose = 'verbose'

    class FakeObject(protocol.message.Message):
        __metaclass__ = python_message.GeneratedProtocolMessageType

        FILE = descriptor.FileDescriptor(__file__, "test", "")
        DESCRIPTOR = descriptor.Descriptor(
            "FakeObject",
            "test.FakeObject",
            filename=__file__,
            file=FILE,
            containing_type=None,
            fields=[
                descriptor.FieldDescriptor(
                    name="name",
                    full_name="test.FakeObject.name",
                    index=0,
                    number=1,
                    type=descriptor.FieldDescriptor.TYPE_STRING,
                    cpp_type=descriptor.FieldDescriptor.CPPTYPE_STRING,
                    label=descriptor.FieldDescriptor.LABEL_REQUIRED,
                    default_value="",
                    message_type=None,
                    enum_type=None,
                    containing_type=None,
                    is_extension=False,
                    extension_scope=None
                ),
                descriptor.FieldDescriptor(
                    name="id",
                    full_name="test.FakeObject.id",
                    index=1,
                    number=2,
                    type=descriptor.FieldDescriptor.TYPE_STRING,
                    cpp_type=descriptor.FieldDescriptor.CPPTYPE_STRING,
                    label=descriptor.FieldDescriptor.LABEL_REQUIRED,
                    default_value="",
                    message_type=None,
                    enum_type=None,
                    containing_type=None,
                    is_extension=False,
                    extension_scope=None
                )
            ], nested_types=[], enum_types=[], extensions=[])

    def makeFakeObject(self):
        returnObj = self.FakeObject()
        returnObj.id = 'id'
        returnObj.name = 'name'
        return returnObj

    def _getRunPrintMethodCalls(self, runner):
        printCalls = []
        with mock.patch('__builtin__.print') as printMethod:
            printMethod.side_effect = \
                lambda *args, **kwargs: printCalls.append((args, kwargs))
            runner.run()
        return printCalls

    def testListReferenceBasesFasta(self):
        args = self.FakeArgs('fasta')
        args.start = 1
        args.end = 100
        returnVal = 'AGCT' * 100  # 400 bases
        runner = cli_client.ListReferenceBasesRunner(args)
        runner._client.list_reference_bases = mock.Mock(
            return_value=returnVal)
        printCalls = self._getRunPrintMethodCalls(runner)
        self.assertEqual(printCalls[0][0][0], '>id:1-100')
        self.assertEqual(len(printCalls), 7)
        self.assertEqual(
            printCalls[-1][0][0],
            returnVal[-50:])  # 50 = 400 % 70

    def testTextOutput(self):
        returnObj = self.makeFakeObject()
        args = self.FakeArgs()
        runner = cli_client.AbstractGetRunner(args)
        runner._method = mock.Mock(return_value=returnObj)
        printCalls = self._getRunPrintMethodCalls(runner)
        self.assertEqual(printCalls, [((u'id', u'name'), {'sep': u'\t'})])

    def testJsonOutput(self):
        returnObj = self.makeFakeObject()
        args = self.FakeArgs('json')
        runner = cli_client.AbstractGetRunner(args)
        runner._method = mock.Mock(return_value=returnObj)
        printCalls = self._getRunPrintMethodCalls(runner)
        self.assertEqual(json.loads(printCalls[0][0][0])['name'], 'name')



"""
Script to imitate a Travis CI run
"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import subprocess

import yaml

import utils
import build_test_data


class TravisSimulator(object):

    logStrPrefix = '***'

    yamlFileLocation = '.travis.yml'

    def parseTestCommands(self):
        yamlFile = file(self.yamlFileLocation)
        yamlData = yaml.load(yamlFile)
        return yamlData['script']

    def runTests(self):
        testCommands = self.parseTestCommands()
        for command in testCommands:
            self.log('Running: "{}"'.format(command))
            subprocess.check_call(command, shell=True)
        self.log('SUCCESS')

    def log(self, logStr):
        utils.log("{0} {1}".format(self.logStrPrefix, logStr))


if __name__ == '__main__':
    build_test_data.buildTestData()
    travisSimulator = TravisSimulator()
    travisSimulator.runTests()



"""
G2P testing on the test data
"""
import unittest

import ga4gh.datamodel as datamodel
import ga4gh.protocol as protocol
import ga4gh.frontend as frontend
import tests.paths as paths


class TestG2P(unittest.TestCase):
    exampleUrl = 'www.example.com'
    phenotypeAssociationSetId = ""

    @classmethod
    def setUpClass(cls):
        config = {
            "DATA_SOURCE": paths.testDataRepo,
            "DEBUG": False
        }
        frontend.reset()
        frontend.configure(
            baseConfig="DevelopmentConfig", extraConfig=config)
        cls.app = frontend.app.test_client()

    def sendSearchRequest(self, path, request, responseClass):
        """
        Sends the specified protocol request instance as JSON, and
        parses the result into an instance of the specified response.
        """
        response = self.sendJsonPostRequest(path, protocol.toJson(request))
        self.assertEqual(200, response.status_code)
        responseData = protocol.fromJson(response.data, responseClass)
        self.assertTrue(
            protocol.validate(protocol.toJson(responseData), responseClass))
        return responseData

    def sendGetRequest(self, path):
        """
        Sends a get request to the specified URL and returns the response.
        """
        return self.app.get(path)

    def getPhenotypeAssociationSetId(self):
        """
        Gets the dataset phenotype association set ID
        """
        request = protocol.SearchDatasetsRequest()
        response = self.sendSearchRequest(
            "datasets/search",
            request,
            protocol.SearchDatasetsResponse)
        datasetId = response.datasets[0].id
        request = protocol.SearchPhenotypeAssociationSetsRequest()
        request.dataset_id = datasetId
        response = self.sendPostRequest(
            "phenotypeassociationsets/search", request)
        response = protocol.fromJson(
            response.data, protocol.SearchPhenotypeAssociationSetsResponse)
        return response.phenotype_association_sets[0].id

    def sendPostRequest(self, path, request):
        """
        Sends the specified GA request object and returns the response.
        """
        headers = {
            'Content-type': 'application/json',
            'Origin': self.exampleUrl,
        }
        return self.app.post(
            path, headers=headers, data=protocol.toJson(request))

    def sendJsonPostRequest(self, path, data):
        """
        Sends a JSON request to the specified path with the specified data
        and returns the response.
        """
        return self.app.post(
            path, headers={'Content-type': 'application/json'},
            data=data)

    def testPhenotypeAssociationSetSearch(self):
        request = protocol.SearchDatasetsRequest()
        response = self.sendSearchRequest(
            "datasets/search",
            request,
            protocol.SearchDatasetsResponse)
        datasetId = response.datasets[0].id
        request = protocol.SearchPhenotypeAssociationSetsRequest()
        request.dataset_id = datasetId
        response = self.sendSearchRequest(
            "phenotypeassociationsets/search",
            request,
            protocol.SearchPhenotypeAssociationSetsResponse)
        # there should be an array
        self.assertIsNotNone(response.phenotype_association_sets)
        # there should be at least one entry
        self.assertGreater(len(response.phenotype_association_sets), 0)

    def getAllDatasets(self):
        """
        Gets all datasets available
        """
        path = 'datasets/search'
        request = protocol.SearchDatasetsRequest()
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchDatasetsResponse)
        return responseData.datasets

    def getAllFeatureSets(self):
        """
        Gets all feature sets available
        """
        datasetId = self.getAllDatasets()[0].id
        datasetName = self.getAllDatasets()[0].name
        path = 'featuresets/search'
        request = protocol.SearchFeatureSetsRequest()
        request.dataset_id = datasetId
        responseData = self.sendSearchRequest(
            path, request, protocol.SearchFeatureSetsResponse)
        return datasetName, responseData.feature_sets

    def getCGDDataSetFeatureSet(self):
        """
        Gets CGD data feature set
        """
        datasetName, featureSets = self.getAllFeatureSets()
        for featureSet in featureSets:
            if featureSet.name == 'cgd':
                return datasetName, featureSet

    def getObfuscatedFeatureCompoundId(
            self, dataSetName, featureSetName, featureId):
        """
        Gets the obfuscated feature compound Id
        """
        splits = [
            dataSetName,
            featureSetName,
            featureId]
        joined = datamodel.FeatureSetCompoundId.join(splits)
        obfuscated = datamodel.FeatureCompoundId.obfuscate(joined)
        return obfuscated

    def testEnsureCGDFeatureSet(self):
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        self.assertIsNotNone(datasetName)
        self.assertIsNotNone(featureSet)
        self.assertIsNotNone(featureSet.name)

    def testEnsureCGDFeatureId(self):
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        featureId = \
            "http://cancer.sanger.ac.uk/cosmic/mutation/overview?id=736"
        obfuscated = self.getObfuscatedFeatureCompoundId(
            datasetName, featureSet.name, featureId)
        compoundId = datamodel.FeatureCompoundId.parse(obfuscated)
        self.assertEqual(featureId, compoundId.featureId)

    def testCompoundFeatureSearch(self):
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        featureId = \
            "http://cancer.sanger.ac.uk/cosmic/mutation/overview?id=736"
        obfuscated = self.getObfuscatedFeatureCompoundId(
            datasetName, featureSet.name, featureId)
        request = protocol.GetFeatureRequest
        request.feature_id = obfuscated
        response = self.sendGetRequest('/features/{}'.format(obfuscated))

        feature = protocol.fromJson(response.data, protocol.Feature)

        self.assertIsNotNone(feature)
        featureId = feature.id

        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.feature_ids.append(featureId)
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(1, len(response.associations))
        self.assertEqual(1, len(response.associations[0].feature_ids))

    def testFeaturesSearchById(self):
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        featureId = \
            "http://cancer.sanger.ac.uk/cosmic/mutation/overview?id=965"
        obfuscated = self.getObfuscatedFeatureCompoundId(
            datasetName, featureSet.name, featureId)
        request = protocol.GetFeatureRequest
        request.feature_id = obfuscated
        response = self.sendGetRequest(
            '/features/{}'.format(obfuscated))

        feature = protocol.fromJson(response.data, protocol.Feature)
        self.assertIsNotNone(feature)
        self.assertEqual(request.feature_id, feature.id)
        self.assertIsNotNone(feature.feature_type)
        self.assertIsNotNone(feature.feature_type.id)
        self.assertEqual(feature.reference_name,  "chr10")
        self.assertEqual(feature.start,  43617416)
        self.assertEqual(feature.end,  43617416)

    def testGenotypesSearchByName(self):
        # setup phenotype query
        request = protocol.SearchFeaturesRequest()
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        request.feature_set_id = featureSet.id
        request.name = "RET M918T missense mutation"

        postUrl = "features/search"
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchFeaturesResponse)
        self.assertEqual(1, len(response.features))
        self.assertEqual(
            "http://cancer.sanger.ac.uk/cosmic/mutation/overview?id=965",
            datamodel.FeatureCompoundId
            .parse(response.features[0].id)
            .featureId
            )
        self.assertEqual(
            request.name,
            response.features[0].name
            )

    def testGenotypesSearchByNameKIT(self):
        request = protocol.SearchFeaturesRequest()
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        request.feature_set_id = featureSet.id
        request.name = \
            "KIT *wild"
        postUrl = "features/search"
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchFeaturesResponse)
        self.assertEqual(3, len(response.features))

    def testPhenotypesSearchById(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        # setup phenotype query
        request.id = "http://ohsu.edu/cgd/30ebfd1a"
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertEqual(request.id, response.phenotypes[0].id)

    def testPhenotypesSearchOntologyTerm(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.type.id = "http://ohsu.edu/cgd/5c895709"
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertGreater(len(response.phenotypes), 0)

    def testPhenotypeSearchQualifiersSensitivity(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        ontologyterm = protocol.OntologyTerm()
        ontologyterm.id = "http://ohsu.edu/cgd/sensitivity"
        request.qualifiers.extend([ontologyterm])
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertGreater(len(response.phenotypes), 0)

    def testPhenotypeSearchQualifiersSensitivityPATO_0000396(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        ontologyterm = protocol.OntologyTerm()
        ontologyterm.id = "http://purl.obolibrary.org/obo/PATO_0000396"
        request.qualifiers.extend([ontologyterm])
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertGreater(len(response.phenotypes), 0)

    def testPhenotypeSearchMultipleQualifiers(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        ontologyterm = protocol.OntologyTerm()
        ontologyterm.id = "http://purl.obolibrary.org/obo/PATO_0000396"
        ontologyterm2 = protocol.OntologyTerm()
        ontologyterm2.id = "http://purl.obolibrary.org/obo/PATO_0000460"
        request.qualifiers.extend([ontologyterm, ontologyterm2])
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertGreater(len(response.phenotypes), 0)

    @unittest.skip
    def testPhenotypesSearchDescription(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.description = \
                "Papillary thyroid carcinoma with sensitivity to therapy"  # noqa
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertGreater(len(response.phenotypes), 0)

    def testPhenotypesSearchDescriptionWildcard(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.description = ".*sensitivity.*"
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertEquals(7, len(response.phenotypes))

    def testPhenotypesSearchMultipleTerms(self):
        request = protocol.SearchPhenotypesRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.description = "Melanoma, NOS with response to therapy"
        request.age_of_onset.id = "http://purl.obolibrary.org/obo/HP_0003581"
        postUrl = '/phenotypes/search'
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchPhenotypesResponse)
        self.assertGreater(len(response.phenotypes), 0)

    def testGenotypePhenotypeSearchFeature(self):
        """
        Search for associations given a feature
        """
        # simulate user interacting with sequenceAnnotations
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        featureId = \
            "http://ohsu.edu/cgd/27d2169c"
        obfuscated = self.getObfuscatedFeatureCompoundId(
            datasetName, featureSet.name, featureId)
        # use the feature to look up associations
        request.feature_ids.extend([obfuscated])
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(1, len(response.associations[0].feature_ids))

    def testGenotypePhenotypeSearchEvidence(self):
        """
        Search for associations given an evidence
        """
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        eq = protocol.EvidenceQuery()
        eq.description = "imatinib"
        request.evidence.extend([eq])
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(1, len(response.associations[0].feature_ids))

    def testGenotypePhenotypeSearchPhenotype(self):
        """
        Search for associations given a phenotype
        """
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.phenotype_ids.extend(["http://ohsu.edu/cgd/25abbb09"])
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(1, len(response.associations[0].feature_ids))

    def testNoFind(self):
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.feature_ids.extend(["FOOBAR"])
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(0, len(response.associations))

    def testGenotypePhenotypeSearchEnsureEvidence(self):
        """
        Ensure evidence level is serialized in responses
        """
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        featureId = \
            "http://ohsu.edu/cgd/27d2169c"
        obfuscated = self.getObfuscatedFeatureCompoundId(datasetName,
                                                         featureSet.name,
                                                         featureId)
        request.feature_ids.extend([obfuscated])
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(1, len(response.associations[0].evidence))
        evidence = response.associations[0].evidence[0]
        self.assertEqual('decreased_sensitivity', evidence.description)

    def testGenotypePhenotypeSearchEnsureEnvironment(self):
        request = protocol.SearchGenotypePhenotypeRequest()
        request.phenotype_association_set_id = \
            self.getPhenotypeAssociationSetId()
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        featureId = \
            "http://ohsu.edu/cgd/27d2169c"
        obfuscated = self.getObfuscatedFeatureCompoundId(
            datasetName, featureSet.name, featureId)
        request.feature_ids.extend([obfuscated])
        eq = protocol.EvidenceQuery()
        eq.description = "imatinib"
        request.evidence.extend([eq])
        response = self.sendSearchRequest(
            '/featurephenotypeassociations/search',
            request,
            protocol.SearchGenotypePhenotypeResponse)
        self.assertEqual(
            1, len(response.associations[0].environmental_contexts))
        environmentalContext = response.associations[0] \
                                       .environmental_contexts[0]
        self.assertEqual('imatinib', environmentalContext.description)

    def _createPagingRequest(self):
        request = protocol.SearchFeaturesRequest()
        datasetName, featureSet = self.getCGDDataSetFeatureSet()
        request.feature_set_id = featureSet.id
        request.name = "KIT *wild"
        return request

    def testGenotypeSearchFeaturePagingOne(self):
        """
        If page size is set to 1 only one association should be returned
        """
        request = self._createPagingRequest()
        request.page_size = 1
        postUrl = "features/search"
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchFeaturesResponse)
        self.assertEqual(1, len(response.features))
        self.assertIsNotNone(response.next_page_token)

    def testGenotypeSearchFeaturePagingMore(self):
        """
        If page size is not set to more than one association should be returned
        """
        request = self._createPagingRequest()
        postUrl = "features/search"
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchFeaturesResponse)
        self.assertGreater(len(response.features), 1)
        self.assertEqual(response.next_page_token, '')

    def testGenotypeSearchFeaturePagingAll(self):
        """
        Loop through all pages
        """
        request = self._createPagingRequest()
        request.page_size = 1
        feature_set_id = request.feature_set_id
        postUrl = "features/search"
        response = self.sendSearchRequest(
            postUrl,
            request,
            protocol.SearchFeaturesResponse)

        self.assertEqual(1, len(response.features))
        self.assertIsNotNone(response.next_page_token)
        pageCount = 1

        while response.next_page_token:
            previous_id = response.features[0].id
            request = protocol.SearchFeaturesRequest()
            request.feature_set_id = feature_set_id
            request.page_size = 1
            request.page_token = response.next_page_token
            request.name = "KIT *wild"
            response = self.sendSearchRequest(
                postUrl,
                request,
                protocol.SearchFeaturesResponse)
            self.assertEqual(1, len(response.features))
            self.assertNotEqual(previous_id, response.features[0].id)
            pageCount += 1
        self.assertEqual(3, pageCount)



"""
Unit tests for sequence annotation Feature and FeatureSet objects.
This is used for all tests that can be performed in isolation
from input data.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.datamodel.sequence_annotations as sequence_annotations
import ga4gh.datamodel.datasets as datasets


class TestAbstractFeatureSet(unittest.TestCase):
    """
    Unit tests for the abstract feature set.
    """
    def setUp(self):
        self._featureSetName = "testFeatureSet"
        self._dataset = datasets.Dataset("test_ds")
        self._featureSet = sequence_annotations.AbstractFeatureSet(
            self._dataset, self._featureSetName)

    def testGetFeatureIdFailsWithNullInput(self):
        self.assertEqual("",
                         self._featureSet.getCompoundIdForFeatureId(None))



"""
Module responsible for translating g2p data into GA4GH native
objects.
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import collections
import rdflib

import ga4gh.datamodel as datamodel
import ga4gh.exceptions as exceptions
import ga4gh.protocol as protocol

# annotation keys
TYPE = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#type'
LABEL = 'http://www.w3.org/2000/01/rdf-schema#label'
HAS_QUALITY = 'http://purl.obolibrary.org/obo/BFO_0000159'


class AbstractPhenotypeAssociationSet(datamodel.DatamodelObject):
    compoundIdClass = datamodel.PhenotypeAssociationSetCompoundId

    def __init__(self, parentContainer, localId):
        super(AbstractPhenotypeAssociationSet, self).__init__(
            parentContainer, localId)

    def toProtocolElement(self):
        pas = protocol.PhenotypeAssociationSet()
        pas.name = self.getLocalId()
        pas.id = self.getId()
        pas.dataset_id = self.getParentContainer().getId()
        return pas


class SimulatedPhenotypeAssociationSet(AbstractPhenotypeAssociationSet):
    def __init__(self, parentContainer, localId, randomSeed,
                 numAssociations=2):
        super(SimulatedPhenotypeAssociationSet, self).__init__(
            parentContainer, localId)
        self._numAssociations = numAssociations

    def getAssociations(self, request=None, featureSets=[]):
        associations = []
        # no request, return a generic set of associations
        if request is None:
            for i in range(self._numAssociations):
                associations.append(self._makeSimulatedAssociation(i))
        # SearchPhenotypesRequest request , return
        # a set of associations that match the request
        elif isinstance(request, protocol.SearchPhenotypesRequest):
            associations.append(self._makeSimulatedAssociation(
                _id=0, phenotype_id=request.id))
        # SearchGenotypePhenotypeRequest request past, return
        # a set of associations that match the request
        else:
            test_phenotype_ids = ["phenotype-0"]
            if len(request.phenotype_ids) > 0:
                test_phenotype_ids = request.phenotype_ids
            for i in range(len(test_phenotype_ids)):
                associations.append(self._makeSimulatedAssociation(
                                    _id=test_phenotype_ids[i].split("-")[1],
                                    phenotype_id=test_phenotype_ids[i]))
        return associations

    def _makeSimulatedAssociation(self, _id=None, phenotype_id=None):
        fpa = protocol.FeaturePhenotypeAssociation()
        fpa.phenotype_association_set_id = self._parentContainer.getId()
        fpa.id = "fpa-{}".format(_id)
        fpa.feature_ids.extend(["feature-{}".format(_id)])
        fpa.evidence.extend([protocol.Evidence()])
        _phenotype_id = "phenotype-{}".format(_id)
        if phenotype_id:
            _phenotype_id = phenotype_id
        fpa.phenotype.MergeFrom(protocol.PhenotypeInstance(
                                id=_phenotype_id))
        fpa.description = "description-{}".format(_id)
        fpa.environmental_contexts.extend(
            [protocol.EnvironmentalContext(id="context-{}".format(_id))])
        return fpa


class G2PUtility(object):

    def _featureTypeLabel(self, featureType):
        """
        Return a label for known types
        """
        featureTypes = {
            'http://purl.obolibrary.org/obo/SO_0001059':
            'sequence_alteration',
            'http://purl.obolibrary.org/obo/SO_0001583':
            'missense_variant',
            'http://purl.obolibrary.org/obo/SO_0000147': 'exon'}
        if featureType in featureTypes:
            return featureTypes[featureType]
        return featureType

    def _extractAssociationsDetails(self, associations):
        """
        Given a set of results from our search query, return the
        `details` (feature,environment,phenotype)
        """
        detailedURIRef = []
        for row in associations.bindings:
            if 'feature' in row:
                detailedURIRef.append(row['feature'])
                detailedURIRef.append(row['environment'])
                detailedURIRef.append(row['phenotype'])
        return detailedURIRef

    def _detailTuples(self, uriRefs):
        """
        Given a list of uriRefs, return a list of dicts:
        {'subject': s, 'predicate': p, 'object': o }
        all values are strings
        """
        details = []
        for uriRef in uriRefs:
            for subject, predicate, object_ in self._rdfGraph.triples(
                    (uriRef, None, None)):
                details.append({
                    'subject': subject.toPython(),
                    'predicate': predicate.toPython(),
                    'object': object_.toPython()
                })
        return details

    def _bindingsToDict(self, bindings):
        """
        Given a binding from the sparql query result,
        create a dict of plain text
        """
        myDict = {}
        for key, val in bindings.iteritems():
            myDict[key.toPython().replace('?', '')] = val.toPython()
        return myDict

    def _addDataFile(self, filename):
        """
        Given a filename, add it to the graph
        """
        if filename.endswith('.ttl'):
            self._rdfGraph.parse(filename, format='n3')
        else:
            self._rdfGraph.parse(filename, format='xml')

    def _getDetails(self, uriRef, associations_details):
        """
        Given a uriRef, return a dict of all the details for that Ref
        use the uriRef as the 'id' of the dict
        """
        associationDetail = {}
        for detail in associations_details:
            if detail['subject'] == uriRef:
                associationDetail[detail['predicate']] = detail['object']
            associationDetail['id'] = uriRef
        return associationDetail

    def _formatExternalIdentifiers(self, element, element_type):
        """
        Formats several external identifiers for query
        """
        elementClause = None
        elements = []
        if not issubclass(element.__class__, dict):
            element = protocol.toJsonDict(element)
        if element['externalIdentifiers']:
            for _id in element['externalIdentifiers']:
                elements.append(self._formatExternalIdentifier(
                    _id, element_type))
            elementClause = "({})".format(" || ".join(elements))
        return elementClause

    def _formatExternalIdentifier(self, element, element_type):
        """
        Formats a single external identifier for query
        """
        if "http" not in element['database']:
            term = "{}:{}".format(element['database'], element['identifier'])
            namespaceTerm = self._toNamespaceURL(term)
        else:
            namespaceTerm = "{}{}".format(
                element['database'], element['identifier'])
        comparison = '?{} = <{}> '.format(element_type, namespaceTerm)
        return comparison

    def _formatOntologyTerm(self, element, element_type):
        """
        Formats the ontology terms for query
        """
        elementClause = None
        if isinstance(element, dict) and element.get('terms'):
            elements = []
            for _term in element['terms']:
                if _term.get('id'):
                    elements.append('?{} = <{}> '.format(
                        element_type, _term['id']))
                else:
                    elements.append('?{} = <{}> '.format(
                        element_type, self._toNamespaceURL(_term['term'])))
            elementClause = "({})".format(" || ".join(elements))
        return elementClause

    def _formatOntologyTermObject(self, terms, element_type):
        """
        Formats the ontology term object for query
        """
        elementClause = None
        if not isinstance(terms, collections.Iterable):
            terms = [terms]
        elements = []
        for term in terms:
            if not issubclass(term.__class__, dict):
                term = protocol.toJsonDict(term)
            if term['id']:
                elements.append('?{} = <{}> '.format(
                    element_type, term['id']))
            else:
                elements.append('?{} = <{}> '.format(
                    element_type, self._toNamespaceURL(term['term'])))
        elementClause = "({})".format(" || ".join(elements))
        return elementClause

    def _formatFeatureClause(self, feature):
        """
        Format a feature for lookup by gene_symbol
        """
        return 'regex(?feature_label, "{}")'.format(feature.gene_symbol)

    def _formatId(self, element, element_type):
        """
        Formats an identifiers for query
        """
        return '?{} = <{}> '.format(
            element_type, element)

    def _formatIds(self, element, element_type):
        """
        Formats a set of identifiers for query
        """
        elementClause = None
        if isinstance(element, collections.Iterable):
            elements = []
            for _id in element:
                elements.append('?{} = <{}> '.format(
                    element_type, _id))
            elementClause = "({})".format(" || ".join(elements))
        return elementClause

    def _formatEvidence(self, elements):
        """
        Formats elements passed into parts of a query for filtering
        """
        elementClause = None
        filters = []
        for evidence in elements:
            if evidence.description:
                elementClause = 'regex(?{}, "{}")'.format(
                    'environment_label', evidence.description)
            if (hasattr(evidence, 'externalIdentifiers') and
                    evidence.externalIdentifiers):
                # TODO will this pick up > 1 externalIdentifiers ?
                for externalIdentifier in evidence['externalIdentifiers']:
                    exid_clause = self._formatExternalIdentifier(
                        externalIdentifier, 'environment')
                    # cleanup parens from _formatExternalIdentifier method
                    elementClause = exid_clause[1:-1]
            if elementClause:
                filters.append(elementClause)
        elementClause = "({})".format(" || ".join(filters))
        return elementClause

    def _baseQuery(self):
        """
        Returns a string with the base of the RDF query structure
        """
        return """
            PREFIX OBAN: <http://purl.org/oban/>
            PREFIX OBO: <http://purl.obolibrary.org/obo/>
            PREFIX dc: <http://purl.org/dc/elements/1.1/>
            PREFIX rdfs: <http://www.w3.org/2000/01/rdf-schema#>
            PREFIX BFO: <http://purl.obolibrary.org/obo/BFO_>
            PREFIX owl: <http://www.w3.org/2002/07/owl#>
            PREFIX faldo: <http://biohackathon.org/resource/faldo#>
            SELECT
                ?association
                ?environment
                ?environment_label
                ?feature
                ?feature_label
                ?phenotype
                ?phenotype_label
                (GROUP_CONCAT(?source; separator="|") AS ?sources)
                ?evidence_type
                ?external_id
                ?phenotype_quality
                WHERE {
                    ?association  a OBAN:association .
                    ?association    OBO:RO_0002558 ?evidence_type .
                    ?association    OBO:RO_has_environment ?environment   .
                    OPTIONAL { ?association  dc:source ?source } .
                    ?association    OBAN:association_has_subject ?feature .
                    ?association    OBAN:association_has_object ?phenotype .
                    ?environment  rdfs:label ?environment_label  .
                    ?phenotype  rdfs:label ?phenotype_label  .
                    ?feature  rdfs:label ?feature_label  .
                    OPTIONAL { ?feature owl:sameAs  ?external_id } .
                    OPTIONAL { ?phenotype BFO:0000159 ?phenotype_quality } .
                    #%FILTER%
                    }
            GROUP  BY ?association
            ORDER  BY ?association
"""

    def _toNamespaceURL(self, term):
        """
        Given an ontologyterm.term return namespace identifier.
        Leverages prefixes already in graph namespace
        Ex.  "DrugBank:DB01268" -> "http://www.drugbank.ca/drugs/DB01268"
        """
        termPrefix, termId = term.split(':')
        for prefix, namespace in self._rdfGraph.namespaces():
            if prefix == termPrefix:
                return "".join([namespace, termId])
        raise exceptions.NotImplementedException(
           "Term has a prefix not found in this instance. {}".format(term))

    def _getIdentifier(self, url):
        """
        Given a url identifier return identifier portion
        Leverages prefixes already in graph namespace
        Returns None if no match
        Ex.  "http://www.drugbank.ca/drugs/DB01268" -> "DB01268"
        """
        for prefix, namespace in self._rdfGraph.namespaces():
            if namespace in url:
                return url.replace(namespace, '')

    def _getPrefix(self, url):
        """
        Given a url return namespace prefix.
        Leverages prefixes already in graph namespace
        Ex.  "http://www.drugbank.ca/drugs/" -> "Drugbank"
        """
        for prefix, namespace in self._rdfGraph.namespaces():
            if namespace.toPython() == url or namespace == url:
                return prefix
        raise exceptions.NotImplementedException(
           "No namespace found for url {}".format(url))

    def _getPrefixURL(self, url):
        """
        Given a url return namespace prefix.
        Leverages prefixes already in graph namespace
        Ex.  "http://www.drugbank.ca/drugs/DDD"
            -> "http://www.drugbank.ca/drugs/"
        """
        for prefix, namespace in self._rdfGraph.namespaces():
            if namespace.toPython() in url:
                return namespace

    def _toGA4GH(self, association, featureSets=[]):
        """
        given an association dict,
        return a protocol.FeaturePhenotypeAssociation
        """

        # The association dict has the keys: environment, environment
        # label, evidence, feature label, phenotype and sources. Each
        # key's value is a dict with the RDF predicates as keys and
        # subject as values

        # 1) map a GA4GH FeaturePhenotypeAssociation
        # from the association dict passed to us
        feature = association['feature']

        fpa = protocol.FeaturePhenotypeAssociation()
        fpa.id = association['id']

        feature_id = feature['id']
        for feature_set in featureSets:
            if self.getLocalId() in feature_set.getLocalId():
                feature_id = feature_set.getCompoundIdForFeatureId(feature_id)

        fpa.feature_ids.extend([feature_id])

        msg = 'Association: genotype:[{}] phenotype:[{}] environment:[{}] ' \
              'evidence:[{}] publications:[{}]'
        fpa.description = msg.format(
            association['feature_label'],
            association['phenotype_label'],
            association['environment_label'],
            self._getIdentifier(association['evidence']),
            association['sources']
            )

        # 2) map a GA4GH Evidence
        # from the association's phenotype & evidence
        evidence = protocol.Evidence()
        phenotype = association['phenotype']

        term = protocol.OntologyTerm()
        term.term = association['evidence_type']
        term.id = phenotype['id']
        term.source_version = self._version
        term.source_name = self._getPrefix(
            self._getPrefixURL(association['id']))
        evidence.evidence_type.MergeFrom(term)

        evidence.description = self._getIdentifier(association['evidence'])

        # 3) Store publications from the list of sources
        for source in association['sources'].split("|"):
            evidence.info['publications'].values.add().string_value = source
        fpa.evidence.extend([evidence])

        # 4) map environment (drug) to environmentalContext
        environmentalContext = protocol.EnvironmentalContext()
        environment = association['environment']
        environmentalContext.id = environment['id']
        environmentalContext.description = association['environment_label']

        term = protocol.OntologyTerm()
        term.term = environment['id']
        term.id = 'http://purl.obolibrary.org/obo/RO_0002606'
        term.source_version = self._version
        term.source_name = self._getPrefix(
            self._getPrefixURL(association['id']))
        environmentalContext.environment_type.MergeFrom(term)

        fpa.environmental_contexts.extend([environmentalContext])

        # 5) map the phenotype
        phenotypeInstance = protocol.PhenotypeInstance()
        term = protocol.OntologyTerm()
        term.term = phenotype[TYPE]
        term.id = phenotype['id']
        term.source_version = self._version
        term.source_name = self._getPrefix(
            self._getPrefixURL(association['id']))
        phenotypeInstance.type.MergeFrom(term)

        phenotypeInstance.description = phenotype[LABEL]
        phenotypeInstance.id = phenotype['id']
        fpa.phenotype.MergeFrom(phenotypeInstance)
        fpa.phenotype_association_set_id = self.getId()
        return fpa


class RdfPhenotypeAssociationSet(G2PUtility, AbstractPhenotypeAssociationSet):
    """
    An rdf object store.  The cancer genome database
    [Clinical Genomics Knowledge Base]
    (http://nif-crawler.neuinfo.org/monarch/ttl/cgd.ttl),
    published by the Monarch project, was the source of Evidence.
    """
    def __init__(self, parentContainer, localId, dataDir):
        """
        Initialize dataset, using the passed dict of sources
        [{source,format}] see rdflib.parse() for more
        If path is set, this backend will load itself
        """
        super(RdfPhenotypeAssociationSet, self).__init__(
            parentContainer, localId)

        # initialize graph
        self._rdfGraph = rdflib.ConjunctiveGraph()
        # save the path
        self._dataUrl = dataDir

        self._scanDataFiles(dataDir, ['*.ttl'])

        # extract version
        cgdTTL = rdflib.URIRef("http://data.monarchinitiative.org/ttl/cgd.ttl")
        versionInfo = rdflib.URIRef(
            u'http://www.w3.org/2002/07/owl#versionInfo')
        self._version = None
        for _, _, obj in self._rdfGraph.triples((cgdTTL, versionInfo, None)):
            self._version = obj.toPython()

    def getAssociations(
            self, request=None, featureSets=[]):
        """
        This query is the main search mechanism.
        It queries the graph for annotations that match the
        AND of [feature,environment,phenotype].
        """
        if len(featureSets) == 0:
            featureSets = self.getParentContainer().getFeatureSets()
        # query to do search
        query = self._formatFilterQuery(request, featureSets)
        associations = self._rdfGraph.query(query)
        # associations is now a dict with rdflib terms with variable and
        # URIrefs or literals

        # given get the details for the feature,phenotype and environment
        associations_details = self._detailTuples(
            self._extractAssociationsDetails(
                associations))

        # association_details is now a list of {subject,predicate,object}
        # for each of the association detail
        # http://nmrml.org/cv/v1.0.rc1/doc/doc/objectproperties/BFO0000159___-324347567.html
        # label "has quality at all times" (en)
        associationList = []
        for assoc in associations.bindings:
            if '?feature' in assoc:
                association = self._bindingsToDict(assoc)
                association['feature'] = self._getDetails(
                    association['feature'],
                    associations_details)
                association['environment'] = self._getDetails(
                    association['environment'],
                    associations_details)
                association['phenotype'] = self._getDetails(
                    association['phenotype'],
                    associations_details)
                association['evidence'] = association['phenotype'][HAS_QUALITY]
                association['id'] = association['association']
                associationList.append(association)

                # our association list is now a list of dicts with the
                # elements of an association: environment, evidence,
                # feature, phenotype and sources, each with their
                # references (labels or URIrefs)

        # create GA4GH objects
        associations = [
            self._toGA4GH(assoc, featureSets) for
            assoc in associationList]
        return associations

    def _formatFilterQuery(self, request=None, featureSets=[]):
        """
        Generate a formatted sparql query with appropriate filters
        """
        query = self._baseQuery()
        filters = []
        if issubclass(request.__class__,
                      protocol.SearchGenotypePhenotypeRequest):
            filters += self._filterSearchGenotypePhenotypeRequest(
                request, featureSets)

        if issubclass(request.__class__, protocol.SearchPhenotypesRequest):
            filters += self._filterSearchPhenotypesRequest(request)

        # apply filters
        filter = "FILTER ({})".format(' && '.join(filters))
        if len(filters) == 0:
            filter = ""
        query = query.replace("#%FILTER%", filter)
        return query

    def _filterSearchGenotypePhenotypeRequest(self, request, featureSets):
        filters = []
        if request.feature_ids:
            featureFilters = []
            for featureId in request.feature_ids:
                for featureSet in featureSets:
                    try:
                        compoundId = datamodel.FeatureCompoundId. \
                                                parse(featureId)
                        # we have a compoundId, so use it to lookup
                        # import ipdb; ipdb.set_trace()
                        if compoundId.feature_set == self.getLocalId():
                            featureFilters.append(
                                self._formatId(
                                    compoundId.featureId, 'feature'))
                            break
                        else:
                            feature = featureSet.getFeature(compoundId)
                            if feature:
                                featureFilters.append(
                                    self._formatFeatureClause(feature))
                                break
                    except Exception:
                        featureFilters.append(
                            self._formatId("NO-FIND", 'feature'))
            if len(featureFilters) > 0:
                filters.append("({})".format(" || ".join(featureFilters)))

        if request.evidence:
            evidenceClause = self._formatEvidence(request.evidence)
            if evidenceClause:
                filters.append(evidenceClause)

        if request.phenotype_ids:
            phenotypeClause = self._formatIds(
                request.phenotype_ids, 'phenotype')
            filters.append(phenotypeClause)

        return filters

    def _filterSearchGenotypesRequest(self, request):
        """
        Filters the request for genotype search requests
        """
        filters = []
        if request.id:
            filters.append("?feature = <{}>".format(request.id))

        if request.reference_name:
            filters.append(
                'regex(?feature_label, "{}")'.format(request.reference_name))

        featureClause = self._formatExternalIdentifiers(request, 'external_id')
        if featureClause:
            filters.append(featureClause)
        # OntologyTerms
        featureOntologytermsClause = self._formatOntologyTerm(
            request, 'feature')
        if featureOntologytermsClause:
            filters.append(featureOntologytermsClause)
        return filters

    def _filterSearchPhenotypesRequest(self, request):
        """
        Filters request for phenotype search requests
        """
        filters = []
        if request.id:
            filters.append("?phenotype = <{}>".format(request.id))

        if request.description:
            filters.append(
                'regex(?phenotype_label, "{}")'.format(request.description))
        # OntologyTerms
        # TODO: refactor this repetitive code
        if hasattr(request.type, 'id') and request.type.id:
            ontolgytermsClause = self._formatOntologyTermObject(
                request.type, 'phenotype')
            if ontolgytermsClause:
                filters.append(ontolgytermsClause)
        if len(request.qualifiers) > 0:
            ontolgytermsClause = self._formatOntologyTermObject(
                request.qualifiers, 'phenotype_quality')
            if ontolgytermsClause:
                filters.append(ontolgytermsClause)
        if hasattr(request.age_of_onset, 'id') and request.age_of_onset.id:
            ontolgytermsClause = self._formatOntologyTermObject(
                request.age_of_onset, 'phenotype_quality')
            if ontolgytermsClause:
                filters.append(ontolgytermsClause)
        return filters



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: ga4gh/rna_quantification.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf.internal import enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
from google.protobuf import descriptor_pb2
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from ga4gh import metadata_pb2 as ga4gh_dot_metadata__pb2


DESCRIPTOR = _descriptor.FileDescriptor(
  name='ga4gh/rna_quantification.proto',
  package='ga4gh',
  syntax='proto3',
  serialized_pb=_b('\n\x1ega4gh/rna_quantification.proto\x12\x05ga4gh\x1a\x14ga4gh/metadata.proto\"D\n\x14RnaQuantificationSet\x12\n\n\x02id\x18\x01 \x01(\t\x12\x12\n\ndataset_id\x18\x02 \x01(\t\x12\x0c\n\x04name\x18\x03 \x01(\t\"\xcf\x01\n\x11RnaQuantification\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x15\n\rbio_sample_id\x18\x08 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x03 \x01(\t\x12\x16\n\x0eread_group_ids\x18\x04 \x03(\t\x12 \n\x08programs\x18\x05 \x03(\x0b\x32\x0e.ga4gh.Program\x12\x17\n\x0f\x66\x65\x61ture_set_ids\x18\x06 \x03(\t\x12!\n\x19rna_quantification_set_id\x18\x07 \x01(\t\"\x8d\x02\n\x0f\x45xpressionLevel\x12\n\n\x02id\x18\x01 \x01(\t\x12\x0c\n\x04name\x18\x02 \x01(\t\x12\x12\n\nfeature_id\x18\x03 \x01(\t\x12\x1d\n\x15rna_quantification_id\x18\x04 \x01(\t\x12\x16\n\x0eraw_read_count\x18\x05 \x01(\x02\x12\x12\n\nexpression\x18\x06 \x01(\x02\x12\x15\n\ris_normalized\x18\x07 \x01(\x08\x12$\n\x05units\x18\x08 \x01(\x0e\x32\x15.ga4gh.ExpressionUnit\x12\r\n\x05score\x18\t \x01(\x02\x12\x19\n\x11\x63onf_interval_low\x18\n \x01(\x02\x12\x1a\n\x12\x63onf_interval_high\x18\x0b \x01(\x02*D\n\x0e\x45xpressionUnit\x12\x1f\n\x1b\x45XPRESSION_UNIT_UNSPECIFIED\x10\x00\x12\x08\n\x04\x46PKM\x10\x01\x12\x07\n\x03TPM\x10\x02\x62\x06proto3')
  ,
  dependencies=[ga4gh_dot_metadata__pb2.DESCRIPTOR,])
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

_EXPRESSIONUNIT = _descriptor.EnumDescriptor(
  name='ExpressionUnit',
  full_name='ga4gh.ExpressionUnit',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='EXPRESSION_UNIT_UNSPECIFIED', index=0, number=0,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='FPKM', index=1, number=1,
      options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='TPM', index=2, number=2,
      options=None,
      type=None),
  ],
  containing_type=None,
  options=None,
  serialized_start=615,
  serialized_end=683,
)
_sym_db.RegisterEnumDescriptor(_EXPRESSIONUNIT)

ExpressionUnit = enum_type_wrapper.EnumTypeWrapper(_EXPRESSIONUNIT)
EXPRESSION_UNIT_UNSPECIFIED = 0
FPKM = 1
TPM = 2



_RNAQUANTIFICATIONSET = _descriptor.Descriptor(
  name='RnaQuantificationSet',
  full_name='ga4gh.RnaQuantificationSet',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ga4gh.RnaQuantificationSet.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='dataset_id', full_name='ga4gh.RnaQuantificationSet.dataset_id', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='ga4gh.RnaQuantificationSet.name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=63,
  serialized_end=131,
)


_RNAQUANTIFICATION = _descriptor.Descriptor(
  name='RnaQuantification',
  full_name='ga4gh.RnaQuantification',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ga4gh.RnaQuantification.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='ga4gh.RnaQuantification.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='bio_sample_id', full_name='ga4gh.RnaQuantification.bio_sample_id', index=2,
      number=8, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='description', full_name='ga4gh.RnaQuantification.description', index=3,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='read_group_ids', full_name='ga4gh.RnaQuantification.read_group_ids', index=4,
      number=4, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='programs', full_name='ga4gh.RnaQuantification.programs', index=5,
      number=5, type=11, cpp_type=10, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='feature_set_ids', full_name='ga4gh.RnaQuantification.feature_set_ids', index=6,
      number=6, type=9, cpp_type=9, label=3,
      has_default_value=False, default_value=[],
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rna_quantification_set_id', full_name='ga4gh.RnaQuantification.rna_quantification_set_id', index=7,
      number=7, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=134,
  serialized_end=341,
)


_EXPRESSIONLEVEL = _descriptor.Descriptor(
  name='ExpressionLevel',
  full_name='ga4gh.ExpressionLevel',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='id', full_name='ga4gh.ExpressionLevel.id', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='name', full_name='ga4gh.ExpressionLevel.name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='feature_id', full_name='ga4gh.ExpressionLevel.feature_id', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='rna_quantification_id', full_name='ga4gh.ExpressionLevel.rna_quantification_id', index=3,
      number=4, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='raw_read_count', full_name='ga4gh.ExpressionLevel.raw_read_count', index=4,
      number=5, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='expression', full_name='ga4gh.ExpressionLevel.expression', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='is_normalized', full_name='ga4gh.ExpressionLevel.is_normalized', index=6,
      number=7, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='units', full_name='ga4gh.ExpressionLevel.units', index=7,
      number=8, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='score', full_name='ga4gh.ExpressionLevel.score', index=8,
      number=9, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conf_interval_low', full_name='ga4gh.ExpressionLevel.conf_interval_low', index=9,
      number=10, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
    _descriptor.FieldDescriptor(
      name='conf_interval_high', full_name='ga4gh.ExpressionLevel.conf_interval_high', index=10,
      number=11, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      options=None),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
  ],
  options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=344,
  serialized_end=613,
)

_RNAQUANTIFICATION.fields_by_name['programs'].message_type = ga4gh_dot_metadata__pb2._PROGRAM
_EXPRESSIONLEVEL.fields_by_name['units'].enum_type = _EXPRESSIONUNIT
DESCRIPTOR.message_types_by_name['RnaQuantificationSet'] = _RNAQUANTIFICATIONSET
DESCRIPTOR.message_types_by_name['RnaQuantification'] = _RNAQUANTIFICATION
DESCRIPTOR.message_types_by_name['ExpressionLevel'] = _EXPRESSIONLEVEL
DESCRIPTOR.enum_types_by_name['ExpressionUnit'] = _EXPRESSIONUNIT

RnaQuantificationSet = _reflection.GeneratedProtocolMessageType('RnaQuantificationSet', (_message.Message,), dict(
  DESCRIPTOR = _RNAQUANTIFICATIONSET,
  __module__ = 'ga4gh.rna_quantification_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.RnaQuantificationSet)
  ))
_sym_db.RegisterMessage(RnaQuantificationSet)

RnaQuantification = _reflection.GeneratedProtocolMessageType('RnaQuantification', (_message.Message,), dict(
  DESCRIPTOR = _RNAQUANTIFICATION,
  __module__ = 'ga4gh.rna_quantification_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.RnaQuantification)
  ))
_sym_db.RegisterMessage(RnaQuantification)

ExpressionLevel = _reflection.GeneratedProtocolMessageType('ExpressionLevel', (_message.Message,), dict(
  DESCRIPTOR = _EXPRESSIONLEVEL,
  __module__ = 'ga4gh.rna_quantification_pb2'
  # @@protoc_insertion_point(class_scope:ga4gh.ExpressionLevel)
  ))
_sym_db.RegisterMessage(ExpressionLevel)


# @@protoc_insertion_point(module_scope)



"""
Tests the biodata module
"""
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import unittest

import ga4gh.datamodel.datasets as datasets
import ga4gh.exceptions as exceptions
import ga4gh.datamodel.bio_metadata as bioMetadata
import ga4gh.protocol as protocol
from ga4gh import pb


class TestIndividuals(unittest.TestCase):
    """
    Tests the Individuals class
    """
    def testToProtocolElement(self):
        dataset = datasets.Dataset('dataset1')
        term = protocol.OntologyTerm()
        term.term = "male genotypic sex"
        term.id = "PATO:0020001"
        term.source_name = "PATO"
        term.source_version = pb.string("2015-11-18")
        # Write out a valid input
        print(protocol.toJsonDict(term))
        validIndividual = protocol.Individual(
            name="test",
            created="2016-05-19T21:00:19Z",
            updated="2016-05-19T21:00:19Z",
            sex=term)
        validIndividual.info['test'].values.add().string_value = 'test-info'
        # pass through protocol creation
        individual = bioMetadata.Individual(
            dataset, "test")
        individual.populateFromJson(protocol.toJson(validIndividual))
        gaIndividual = individual.toProtocolElement()
        # Verify elements exist
        self.assertEqual(gaIndividual.created, validIndividual.created)
        self.assertEqual(gaIndividual.updated, validIndividual.updated)
        # Invalid input
        invalidIndividual = '{"bad:", "json"}'
        individual = bioMetadata.Individual(dataset, "test")
        # Should fail
        self.assertRaises(
            exceptions.InvalidJsonException,
            individual.populateFromJson,
            invalidIndividual)


class TestBioSamples(unittest.TestCase):
    """
    Tests the BioSamples class
    """
    def testToProtocolElement(self):
        dataset = datasets.Dataset('dataset1')
        # Write out a valid input
        validBioSample = protocol.BioSample(
            name="test",
            created="2016-05-19T21:00:19Z",
            updated="2016-05-19T21:00:19Z")
        validBioSample.info['test'].values.add().string_value = 'test-info'
        # pass through protocol creation
        bioSample = bioMetadata.BioSample(
            dataset, "test")
        bioSample.populateFromJson(protocol.toJson(validBioSample))
        gaBioSample = bioSample.toProtocolElement()
        # Verify elements exist
        self.assertEqual(gaBioSample.created, validBioSample.created)
        self.assertEqual(gaBioSample.updated, validBioSample.updated)
        # Invalid input
        invalidBioSample = '{"bad:", "json"}'
        bioSample = bioMetadata.Individual(dataset, "test")
        # Should fail
        self.assertRaises(
            exceptions.InvalidJsonException,
            bioSample.populateFromJson,
            invalidBioSample)
